# coding: utf-8
#
# borrowed from
#   https://github.com/jic-dtool/dtool-lookup-server/blob/master/tests/__init__.py
# 05 Nov 2020
#
"""Fixtures common utility functions."""

import logging
import random
import socket
import string


def _random_string(
        size=9,
        prefix="test_",
        chars=string.ascii_uppercase + string.ascii_lowercase + string.digits):
    return prefix + ''.join(random.choice(chars) for _ in range(size))


def _allocate_random_free_port(host='localhost'):
    logger = logging.getLogger(__name__)
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind((host, 0))
    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    _, local_port = s.getsockname()
    logger.info("Allocated free local port: {:d}".format(local_port))
    return local_port



# coding: utf-8
#
# multiprocessing.py
#
# Copyright (C) 2020 IMTEK Simulation
# Author: Johannes Hoermann, johannes.hoermann@imtek.uni-freiburg.de
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Extensions to fireworks dataflow tasks."""
import copy
import glob
import io
import logging
import os

from typing import Dict, List

from abc import abstractmethod
from contextlib import ExitStack

from fireworks.core.firework import Firework, FireTaskBase, FWAction, Workflow
from fireworks.fw_config import FW_LOGGING_FORMAT
from fireworks.utilities.fw_serializers import load_object, ENCODING_PARAMS
from fireworks.utilities.dict_mods import get_nested_dict_value, set_nested_dict_value

from jlhfw.utils.dict import compare, dict_merge, apply_mod_spec
from jlhfw.utils.logging import LoggingContext, _log_nested_dict
from jlhfw.utils.multiprocessing import RunAsChildProcessTask

__author__ = 'Johannes Laurin Hoermann'
__copyright__ = 'Copyright 2020, IMTEK Simulation, University of Freiburg'
__email__ = 'johannes.hoermann@imtek.uni-freiburg.de, johannes.laurin@gmail.com'
__date__ = 'Nov 28, 2020'

DEFAULT_FORMATTER = logging.Formatter(FW_LOGGING_FORMAT)


def from_fw_spec(param, fw_spec):
    """Expands param['key'] as key within fw_spec.

    If param is dict hand has field 'key', then return value at specified
    position from fw_spec. Otherwise, return 'param' itself.
    """
    if isinstance(param, dict) and 'key' in param:
        ret = get_nested_dict_value(fw_spec, param['key'])
    else:
        ret = param
    return ret


# TODO: This follows almost the same pattern as DtoolTask, further abstraction possible
class DataflowTask(RunAsChildProcessTask):
    """
    A dataflow task ABC.

    Required params:
        None
    Optional params:
        - output_key (str): spec key that will be used to pass a task's output
            to child fireworks. Default: None
        - dict_mod (str, default: '_set'): how to insert handled dataset's
            properties into output key, see fireworks.utils.dict_mods
        - propagate (bool, default:None): if True, then set the
            FWAction 'propagate' flag and propagate updated fw_spec not only to
            direct children, but to all descendants down to wokflow's leaves.
        - stored_data (bool, default: False): put handled dataset properties
            into FWAction.stored_data
        - store_stdlog (bool, default: False): insert log output into FWAction.stored_data
        - stdlog_file (str, Default: NameOfTaskClass.log): print log to file
        - loglevel (str, Default: logging.INFO): loglevel for this task
    """
    _fw_name = 'DataflowTask'
    required_params = [*RunAsChildProcessTask.required_params]
    optional_params = [
        *RunAsChildProcessTask.optional_params,
        "stored_data",
        "output_key",
        "dict_mod",
        "propagate",
        "stdlog_file",
        "store_stdlog",
        "loglevel"]

    @abstractmethod
    def _run_task_internal(self, fw_spec) -> List[Dict]:
        """Derivatives implement their functionality here."""
        ...

    def _run_task_as_child_process(self, fw_spec, q, e=None):
        """q is a Queue used to return fw_action."""
        stored_data = self.get('stored_data', False)
        output_key = self.get('output_key', None)
        dict_mod = self.get('dict_mod', '_set')
        propagate = self.get('propagate', False)

        stdlog_file = self.get('stdlog_file', '{}.log'.format(self._fw_name))
        store_stdlog = self.get('store_stdlog', False)

        loglevel = self.get('loglevel', logging.INFO)

        with ExitStack() as stack:

            if store_stdlog:
                stdlog_stream = io.StringIO()
                logh = logging.StreamHandler(stdlog_stream)
                logh.setFormatter(DEFAULT_FORMATTER)
                stack.enter_context(
                    LoggingContext(handler=logh, level=loglevel, close=False))

            # logging to dedicated log file if desired
            if stdlog_file:
                logfh = logging.FileHandler(stdlog_file, mode='a', **ENCODING_PARAMS)
                logfh.setFormatter(DEFAULT_FORMATTER)
                stack.enter_context(
                    LoggingContext(handler=logfh, level=loglevel, close=True))

            output = self._run_task_internal(fw_spec)

        if isinstance(output, FWAction):
            fw_action = output
        else:
            fw_action = FWAction()

            if stored_data:
                fw_action.stored_data = {}
                fw_action.stored_data['output'] = output

            if output_key:  # inject into fw_spec
                fw_action.mod_spec = [{dict_mod: {output_key: output}}]

        # ATTENTION: might override stored_data, stdlog set in task
        if store_stdlog and fw_action.stored_data is None:
            fw_action.stored_data = {}
        if store_stdlog:
            stdlog_stream.flush()
            fw_action.stored_data['stdlog'] = stdlog_stream.getvalue()

        # 'propagate' only development feature for now
        if hasattr(fw_action, 'propagate') and propagate:
            fw_action.propagate = propagate

        # return fw_action
        q.put(fw_action)


class SearchDictTask(DataflowTask):
    """
    Search for 'search' within 'input' and return list of matching keys (or indices in case of list) as 'output'.

    Required params:
        None
    Optional params:
        - input (dict or list): dict or list to search
        = input_key (str): if specified, then supersedes 'input' by entry in fw_spec this key points to.
            One of 'input' and 'input_key' must be specified.
        - search (obj): object to search for in 'input'.
        - search_key (str): if specified, then supersedes 'search' by entry in fw_spec this key points to.
            One of 'search' and 'search_key' must be specified.
        - marker (dict or list):  if specified, must mimic structure of entries in 'input' and mark fields for.
            comparison with boolean values. If None (default), then compare everything.
        - marker_key (str):  if specified, then supersedes 'marker' by entry in fw_spec this key points to.
            One of 'marker' and 'marker_key' must be specified.
        - limit (int): limit the number of results. If None, then no limit (default).
        - expand (bool): will replace list result with single value result if list result only contains one entry and
            with None if list result is empty. Default: False.

    Fields 'limit', 'expand_one' may also be a dict of format { 'key': 'some->nested->fw_spec->key' }
    for looking up value within fw_spec instead.
    """
    _fw_name = 'SearchDiskTask'
    required_params = [*DataflowTask.required_params]
    optional_params = [
        *DataflowTask.optional_params,
        "input",  # dict or list to search
        "input_key",
        "search",  # entry to search for
        "search_key",
        "marker",  # marker must mimic structure of entries in input and mark fields for comparison with boolean values.
        "marker_key",
        "limit",
        "expand",
    ]

    def _run_task_internal(self, fw_spec):
        logger = logging.getLogger(__name__)

        input = self.get('input', None)
        input_key = self.get('input_key', None)

        search = self.get('search', None)
        search_key = self.get('search_key', None)

        marker = self.get('marker', None)
        marker_key = self.get('marker_key', None)

        limit = self.get('limit', None)
        limit = from_fw_spec(limit, fw_spec)

        expand = self.get('expand', None)
        expand = from_fw_spec(expand, fw_spec)

        if input_key:
            logger.debug("input from fw_spec at '%s'." % input_key)
            input = get_nested_dict_value(fw_spec, input_key)
        elif input:
            pass
        else:
            raise ValueError("Neither 'input' nor 'input_key' specified.")

        if search_key:
            logger.debug("search from fw_spec at '%s'." % search_key)
            search = get_nested_dict_value(fw_spec, search_key)
        elif input:
            pass
        else:
            raise ValueError("Neither 'search' nor 'search_key' specified.")

        if marker_key:
            logger.debug("marker from fw_spec at '%s'." % marker_key)
            marker = get_nested_dict_value(fw_spec, marker_key)
        elif input:
            pass
        else:
            logger.warning("Neither 'marker' nor 'marker_key' specified, compare everything.")

        logger.debug("input:")
        _log_nested_dict(logger.debug, input)

        logger.debug("search:")
        _log_nested_dict(logger.debug, search)

        logger.debug("marker:")
        _log_nested_dict(logger.debug, marker)

        matches = []
        def find_match(key, entry):
            if compare(entry, search, marker):
                logger.info("Found match at %s: %s" % (key, entry))
                matches.append(key)

        if isinstance(input, dict):
            for key, entry in input.items():
                find_match(key, entry)
        elif isinstance(input, list):
            for key, entry in enumerate(input):
                find_match(key, entry)
        else:
            ValueError("type of 'input' is '%s', but must be 'dict' or 'list'." % type(input))

        logger.info("Found matches at '%s'" % matches)

        if isinstance(limit, int) and limit >=0 :
            matches = matches[:limit]
            logger.debug("Limit matches to first %d: '%s'" % (limit, matches))

        if expand and len(matches) == 1:
            matches = matches[0]
            logger.debug("Expand single-entry result'%s'." % matches)
        elif expand and len(matches) == 0:
            matches = None
            logger.debug("Expand empty result as None.")

        logger.info("Return '%s'" % matches)
        return matches


# Overrides for default Fireworks dataflow tasks

class ExtendedForeachTask(DataflowTask):
    """
    This firetask branches the workflow creating parallel fireworks
    using FWAction: one firework for each element or each chunk from the
    *split* list. Each firework in this generated list contains the firetask
    specified in the *task* dictionary. If the number of chunks is specified
    the *split* list will be divided into this number of chunks and each
    chunk will be processed by one of the generated child fireworks.

    Required params:
        - task (dict): a dictionary version of the firetask
        - split (str or [str]): label  an input list or a list of such;
          they must be available both in
          the *inputs* list of the specified task and in the spec.

    Optional params:
        - number of chunks (int): if provided the *split* input list will be
          divided into this number of sublists and each will be processed by
          a separate child firework
    """
    _fw_name = 'ExtendedForeachTask'
    required_params = [
        *DataflowTask.required_params,
        "task",
        "split"]
    optional_params = [
        *DataflowTask.optional_params,
        "number of chunks"
    ]

    def _run_task_internal(self, fw_spec):
        assert isinstance(self['split'], (str,list)), self['split']
        split_list = self['split']
        if isinstance( split_list, str): split_list = [split_list]

        reflen = 0
        for split in split_list:
            assert isinstance(fw_spec[split], list)
            #if isinstance(self['task']['inputs'], list):
            #    assert split in self['task']['inputs']
            #else: # only one inputs entry , str
            #    assert split == self['task']['inputs']

            split_field = fw_spec[split]
            lensplit = len(split_field)

            # update reflen on first iteration
            if reflen == 0:
                assert lensplit != 0, ('input to split is empty:', split)
                reflen = lensplit
                nchunks = self.get('number of chunks')
                if not nchunks:
                    nchunks = lensplit
                chunklen = lensplit // nchunks
                if lensplit % nchunks > 0:
                    chunklen = chunklen + 1

                chunks = [ { split: split_field[i:i+chunklen] } for i in range(0, lensplit, chunklen)]
            else:
                assert lensplit == reflen, ('input lists not of equal length:', split)
                for i in range(0, lensplit, chunklen):
                    chunks[i//chunklen].update( { split: split_field[i:i+chunklen] } )

        fireworks = []
        chunk_index_spec = self.get('chunk index spec')

        # allow for multiple tasks
        task_list = self['task']
        if not isinstance( task_list, list ):
            task_list = [ task_list ]
        for index, chunk in enumerate(chunks):
            spec = fw_spec.copy()
            for split in split_list:
                spec[split] = chunk[split]

            tasks = []
            for task_entry in task_list:
                task = load_object(task_entry)
                task['chunk_number'] = index
                tasks.append(task)

            if chunk_index_spec and isinstance(chunk_index_spec, str):
                spec[chunk_index_spec] = index
            name = self._fw_name + ' ' + str(index)
            fireworks.append(Firework(tasks, spec=spec, name=name))
        return FWAction(detours=fireworks)


class BranchWorkflowTask(DataflowTask):
    """
    Similarly to the ForeachTask, this task branches the workflow creating
    parallel fireworks using FWAction: instead of one firework for each element
    or each chunk from the *split* list, however, arbitrary subworkflows of
    multiple fireworks may be appended as additions or detours for each chunk.
    If the number of chunks is specified
    the *split* list will be divided into this number of chunks and each
    chunk will be processed by one of the generated child fireworks.

    Required parameters:
        - split (str or [str]): label  an input list or a list of such;
          they must be available both in
          the *inputs* list of the specified task and in the spec.

    Optional parameters:
        - number of chunks (int): if provided the *split* input list will be
          divided into this number of sublists and each will be processed by
          a separate child firework

        - addition_wf (dict): Workflow or single FireWork to always append as
            an addition. Default: None.
        - detour_wf (dict): Workflow or single FireWork to always append as
            a detour. Default: None

        NOTE: at least one of addition_wf and detour_wf must be specified.

        - detour_fws_root ([int]): fw_ids (referring to fws in detour_wf)
            to identify "roots" to be connected to this fw. If not
            specified, all dangling roots are connected.
        - detour_fws_leaf ([int]): fw_ids (referring to fws in detour_wf)
            to identify "leaves" to be connected to this fw's direct children.
            If not specified, all dangling leaves are connected.
        - addition_fws_root ([int]): fw_ids (referring to fws in addition_wf)
            to identify "roots" to be connected to this fw. If not
            specified, all dangling roots are connected.

        - apply_mod_spec_to_addition_wf (bool): Apply FWAction's update_spec and
            mod_spec to 'addition_wf', same as for all other regular childern
            of this task's FireWork. Default: True.
        - apply_mod_spec_to_detour_wf (bool): Apply FWAction's update_spec and
            mod_spec to 'detour_wf', , same as for all other regular children
            of this task's FireWork. Default: True.
        - superpose_addition_on_my_fw_spec (bool):
            merge own fw_spec with fw_spec of all FireWorks within addition_wf,
            with latter enjoying precedence. Default: False.
        - superpose_detour_on_my_fw_spec (bool):
            Merge own fw_spec with fw_spec of all FireWorks within detour_wf,
            with latter enjoying precedence. Default: False.
        - addition_fw_spec_to_exclude ([str]):
        - detour_fw_spec_to_exclude ([str]):
            When any of the above superpose flags is set, the top-level fw_spec
            fields specified here won't enter the created additions.
            Default for are all reserved fields, i.e. [
                '_add_fworker',
                '_add_launchpad_and_fw_id',
                '_allow_fizzled_parents',
                '_background_tasks',
                '_category',
                '_dupefinder',
                '_files_in',
                '_files_out',
                '_files_prev',
                '_fizzled_parents',
                '_fw_env',
                '_fworker',
                '_job_info',
                '_launch_dir',
                '_pass_job_info',
                '_preserve_fworker',
                '_priority',
                '_queueadapter',
                '_tasks',
                '_trackers',
            ]

        - output (str): spec key that will be used to pass output to child
            fireworks. Default: None
        - dict_mod (str, default: '_set'): how to insert output into output
            key, see fireworks.utils.dict_mods
        - propagate (bool, default: None): if True, then set the
            FWAction 'propagate' flag and propagate updated fw_spec not only to
            direct children, but to all descendants down to wokflow's leaves.
        - stored_data (bool, default: False): put outputs into database via
            FWAction.stored_data
        - store_stdlog (bool, default: False): insert log output into database
            (only if 'stored_data' or 'output' is spcified)
        - stdlog_file (str, Default: NameOfTaskClass.log): print log to file
        - loglevel (str, Default: logging.INFO): loglevel for this task

    Fields  'ignore_errors'
    may also be a dict of format { 'key': 'some->nested->fw_spec->key' } for
    looking up value within 'fw_spec' instead.

    NOTE: reserved fw_spec keywords are (alphabetically)
        - _add_fworker
        - _add_launchpad_and_fw_id
        - _allow_fizzled_parents
        - _background_tasks
        - _category
        - _dupefinder
        - _files_in
        - _files_out
        - _files_prev
        - _fizzled_parents
        - _fw_env
        - _fworker
        - _job_info
        - _launch_dir
        - _pass_job_info
        - _preserve_fworker
        - _priority
        - _queueadapter
        - _tasks
        - _trackers
    """
    _fw_name = 'BranchWorkflowTask'
    required_params = [
        *DataflowTask.required_params,
        "split"]
    optional_params = [
        *DataflowTask.optional_params,
        "number of chunks",

        "detour_wf",
        "addition_wf",
        "detour_fws_root",
        "detour_fws_leaf",
        "addition_fws_root",
        "apply_mod_spec_to_addition_wf",
        "apply_mod_spec_to_detour_wf",

        "superpose_addition_on_my_fw_spec",
        "superpose_detour_on_my_fw_spec",
        "addition_fw_spec_to_exclude",
        "detour_fw_spec_to_exclude",
    ]

    def appendable_wf_from_dict(self, obj_dict, base_spec=None, exclusions={}):
        """Creates Workflow from a Workflow or single FireWork dict description.

        If specified, use base_spec for all fw_spec and superpose individual
        specs on top.

        Args:
            - obj_dict (dict): describes either single FW or whole Workflow
            - base_spec (dict): use those specs for all FWs within workflow.
                Specific specs already set within obj_dict take precedence.
            - exclusions (dict): nested dict with keys marked for exclusion
                by True boolean value. Excluded keys are stripped off base_spec.

        Returns:
            (Workflow, dict) tuple
            - Worfklow: created workflow
            - {int: int}: old to new fw_ids mapping
        """
        logger = logging.getLogger(__name__)

        logger.debug("Initial obj_dict:")
        _log_nested_dict(logger.debug, obj_dict)

        if base_spec:
            logger.debug("base_spec:")
            _log_nested_dict(logger.debug, base_spec)

        if exclusions:
            logger.debug("exclusions:")
            _log_nested_dict(logger.debug, exclusions)

        remapped_fw_ids = {}
        if isinstance(obj_dict, dict):
            # in case of single Fireworks:
            if "spec" in obj_dict:
                # append firework (defined as dict):
                if base_spec:
                    obj_dict["spec"] = dict_merge(base_spec, obj_dict["spec"],
                                                  exclusions=exclusions)
                fw = Firework.from_dict(obj_dict)
                remapped_fw_ids[fw.fw_id] = self.consecutive_fw_id
                fw.fw_id = self.consecutive_fw_id
                self.consecutive_fw_id -= 1
                wf = Workflow([fw])
            else:   # if no single fw, then wf
                if base_spec:
                    for fw_dict in obj_dict["fws"]:
                        fw_dict["spec"] = dict_merge(base_spec, fw_dict["spec"],
                                                     exclusions=exclusions)
                wf = Workflow.from_dict(obj_dict)
                # do we have to reassign fw_ids? yes
                for fw in wf.fws:
                    remapped_fw_ids[fw.fw_id] = self.consecutive_fw_id
                    fw.fw_id = self.consecutive_fw_id
                    self.consecutive_fw_id -= 1
                wf._reassign_ids(remapped_fw_ids)
        else:
            raise ValueError("type({}) is '{}', but 'dict' expected.".format(
                             obj_dict, type(obj_dict)))
        logger.debug("Built object:")
        _log_nested_dict(logger.debug, wf.as_dict())

        return wf, remapped_fw_ids

    # if the curret fw yields outfiles, then check whether according
    # '_files_prev' must be written for newly created insertions
    def write_files_prev(self, wf, fw_spec, root_fw_ids=None):
        "Sets _files_prev in roots of new workflow according to _files_out in fw_spec."
        logger = logging.getLogger(__name__)

        if fw_spec.get("_files_out"):
            logger.info("Current FireWork's '_files_out': {}".format(
                        fw_spec.get("_files_out")))

            files_prev = {}

            for k, v in fw_spec.get("_files_out").items():
                files = glob.glob(os.path.join(os.curdir, v))
                if files:
                    logger.info("This Firework provides {}: {}".format(
                                k, files), " within _files_out.")
                    filepath = os.path.abspath(sorted(files)[-1])
                    logger.info("{}: '{}' provided as '_files_prev'".format(
                                k, filepath), " to subsequent FireWorks.")
                    files_prev[k] = filepath

            # get roots of insertion wf and assign _files_prev to them
            if root_fw_ids is None:
                root_fw_ids = wf.root_fw_ids
            root_fws = [fw for fw in wf.fws if fw.fw_id in root_fw_ids]

            for root_fw in root_fws:
                root_fw.spec["_files_prev"] = files_prev

        return wf

    def _run_task_internal(self, fw_spec):
        logger = logging.getLogger(__name__)

        # from original ForeachTask
        assert isinstance(self['split'], (str, list)), self['split']
        split_list = self['split']
        if isinstance(split_list, str): split_list = [split_list]
        logger.debug("Iteratring through '{}'.", split_list)
        reflen = 0
        for split in split_list:
            logger.debug("Splitting field '{}'.", split)
            assert isinstance(fw_spec[split], list)
            # if isinstance(self['task']['inputs'], list):
            #    assert split in self['task']['inputs']
            # else: # only one inputs entry , str
            #    assert split == self['task']['inputs']

            split_field = fw_spec[split]
            logger.debug("Field content '{}'.".format(split_field))
            lensplit = len(split_field)
            logger.debug("Length {}.".format(lensplit))

            # update reflen on first iteration
            if reflen == 0:
                assert lensplit != 0, ('input to split is empty:', split)
                reflen = lensplit
                nchunks = self.get('number of chunks')
                if not nchunks:
                    nchunks = lensplit
                chunklen = lensplit // nchunks
                if lensplit % nchunks > 0:
                    chunklen = chunklen + 1

                chunks = [{split: split_field[i:i + chunklen]} for i in range(0, lensplit, chunklen)]
            else:
                assert lensplit == reflen, ('input lists not of equal length:', split)
                for i in range(0, lensplit, chunklen):
                    chunks[i // chunklen].update({split: split_field[i:i + chunklen]})

            logger.debug("Split into {} chunks of length {}: {}.".format(nchunks, chunklen, chunks))

        chunk_index_spec = self.get('chunk index spec')

        # following adaptedfrom RecoverTask

        # NOTE: be careful to distinguish between what is referred to as
        # detour_wf in the task's docstring, stored within detour_wf_dict below
        # and the final detour_wf constructed, possibly comprising the partial
        # restart and detour workflows specified via task parameters as well as
        # another copy of this recover_fw.

        self.consecutive_fw_id = -1  # quite an ugly necessity
        # get fw_spec entries or their default values:
        detour_wf_dict = self.get('detour_wf', None)
        addition_wf_dict = self.get('addition_wf', None)

        detour_fws_root = self.get('detour_fws_root', None)
        detour_fws_leaf = self.get('detour_fws_leaf', None)
        addition_fws_root = self.get('addition_fws_root', None)

        apply_mod_spec_to_addition_wf = self.get('apply_mod_spec_to_addition_wf', True)
        apply_mod_spec_to_addition_wf = from_fw_spec(apply_mod_spec_to_addition_wf,
                                                     fw_spec)

        apply_mod_spec_to_detour_wf = self.get('apply_mod_spec_to_detour_wf', True)
        apply_mod_spec_to_detour_wf = from_fw_spec(apply_mod_spec_to_detour_wf,
                                                   fw_spec)

        superpose_addition_on_my_fw_spec = self.get(
            'superpose_addition_on_my_fw_spec', False)
        superpose_addition_on_my_fw_spec = from_fw_spec(
            superpose_addition_on_my_fw_spec, fw_spec)

        superpose_detour_on_my_fw_spec = self.get(
            'superpose_detour_on_my_fw_spec', False)
        superpose_detour_on_my_fw_spec = from_fw_spec(
            superpose_detour_on_my_fw_spec, fw_spec)

        default_fw_spec_to_exclude = [
            '_add_fworker',
            '_add_launchpad_and_fw_id',
            '_allow_fizzled_parents',
            '_background_tasks',
            '_category',
            '_dupefinder',
            '_files_in',
            '_files_out',
            '_files_prev',
            '_fizzled_parents',
            '_fw_env',
            '_fworker',
            '_job_info',
            '_launch_dir',
            '_pass_job_info',
            '_preserve_fworker',
            '_priority',
            '_queueadapter',
            '_tasks',
            '_trackers',
        ]

        addition_fw_spec_to_exclude = self.get('addition_fw_spec_to_exclude', default_fw_spec_to_exclude)
        if isinstance(addition_fw_spec_to_exclude, list):
            addition_fw_spec_to_exclude_dict = {k: True for k in addition_fw_spec_to_exclude}
        else:  # supposed to be dict then
            addition_fw_spec_to_exclude_dict = addition_fw_spec_to_exclude

        detour_fw_spec_to_exclude = self.get('detour_fw_spec_to_exclude', default_fw_spec_to_exclude)
        if isinstance(detour_fw_spec_to_exclude, list):
            detour_fw_spec_to_exclude_dict = {k: True for k in detour_fw_spec_to_exclude}
        else:  # supposed to be dict then
            detour_fw_spec_to_exclude_dict = detour_fw_spec_to_exclude

        # input assertions, ATTENTION: order matters

        # find other files to forward:
        file_list = []

        # distinguish between FireWorks and Workflows by top-level keys
        # fw: ['spec', 'fw_id', 'created_on', 'updated_on', 'name']
        # wf: ['fws', 'links', 'name', 'metadata', 'updated_on', 'created_on']
        detour_wf_list = []
        addition_wf_list = []

        all_mapped_detour_fws_root = []
        all_mapped_detour_fws_leaf = []
        all_mapped_addition_fws_root = []

        fw_action = FWAction()

        for index, chunk in enumerate(chunks):
            minimal_base_spec = {}
            full_base_spec = copy.deepcopy(fw_spec)
            for split in split_list:
                minimal_base_spec[split] = chunk[split]
                full_base_spec[split] = chunk[split]

            minimal_base_spec['chunk_number'] = index
            full_base_spec['chunk_number'] = index

            if chunk_index_spec and isinstance(chunk_index_spec, str):
                minimal_base_spec[chunk_index_spec] = index
                full_base_spec[chunk_index_spec] = index

            detour_wf = None
            addition_wf = None
            mapped_detour_fws_root = []
            mapped_detour_fws_leaf = []
            mapped_addition_fws_root = []

            # build detour wf
            if isinstance(detour_wf_dict, dict):
                new_detour_wf_dict = copy.deepcopy(detour_wf_dict)
                if superpose_detour_on_my_fw_spec:
                    detour_wf_base_spec = copy.deepcopy(full_base_spec)
                else:
                    detour_wf_base_spec = copy.deepcopy(minimal_base_spec)

                logger.debug("Base spec for detour {}:".format(index))
                _log_nested_dict(logger.debug, detour_wf_base_spec)

                detour_wf, detour_wf_fw_id_mapping = self.appendable_wf_from_dict(
                    new_detour_wf_dict, base_spec=detour_wf_base_spec,
                    exclusions=detour_fw_spec_to_exclude_dict)

                if detour_fws_root is None:  # default, as in core fireworks
                    mapped_detour_fws_root.extend(detour_wf.root_fw_ids)
                elif isinstance(detour_fws_root, (list, tuple)):
                    mapped_detour_fws_root.extend(
                        [detour_wf_fw_id_mapping[fw_id] for fw_id in detour_fws_root])
                else:  # isinstance(detour_fws_root, int)
                    mapped_detour_fws_root.append(detour_wf_fw_id_mapping[detour_fws_root])

                if detour_fws_leaf is None:  # default, as in core fireworks
                    mapped_detour_fws_leaf.extend(detour_wf.leaf_fw_ids)
                elif isinstance(detour_fws_leaf, (list, tuple)):
                    mapped_detour_fws_leaf.extend(
                        [detour_wf_fw_id_mapping[fw_id] for fw_id in detour_fws_leaf])
                else:  # isinstance(detour_fws_leaf, int)
                    mapped_detour_fws_leaf.append(detour_wf_fw_id_mapping[detour_fws_leaf])

                # only log if sepcific roots or leaves had been specified
                if detour_fws_root:
                    logger.debug("Mapped detour_wf root fw_ids {} onto newly created fw_ids {}".format(
                        detour_fws_root, mapped_detour_fws_root[-len(detour_fws_root):]))
                if detour_fws_leaf:
                    logger.debug("Mapped detour_wf leaf fw_ids {} onto newly created fw_ids {}".format(
                        detour_fws_leaf, mapped_detour_fws_leaf[-len(detour_fws_leaf)]))

                if apply_mod_spec_to_detour_wf:
                    apply_mod_spec(detour_wf, fw_action, fw_ids=mapped_detour_fws_root)

                self.write_files_prev(detour_wf, fw_spec, root_fw_ids=mapped_detour_fws_root)
                logger.debug("detour_wf:")
                _log_nested_dict(logger.debug, detour_wf.as_dict())

                detour_wf_list.append(detour_wf)
                all_mapped_detour_fws_root.append(mapped_detour_fws_root)
                all_mapped_detour_fws_leaf.append(mapped_detour_fws_leaf)

            # addition wf
            if isinstance(addition_wf_dict, dict):
                new_addition_wf_dict = copy.deepcopy(addition_wf_dict)
                if superpose_addition_on_my_fw_spec:
                    addition_wf_base_spec = copy.deepcopy(full_base_spec)
                else:
                    addition_wf_base_spec = copy.deepcopy(minimal_base_spec)

                logger.debug("Base spec for addition {}:".format(index))
                _log_nested_dict(logger.debug, addition_wf_base_spec)

                addition_wf, addition_wf_fw_id_mapping = self.appendable_wf_from_dict(
                    new_addition_wf_dict, base_spec=addition_wf_base_spec,
                    exclusions=addition_fw_spec_to_exclude_dict)

                if addition_fws_root is None:
                    mapped_addition_fws_root.extend(addition_wf.root_fw_ids)
                elif isinstance(addition_fws_root, (list, tuple)):
                    mapped_addition_fws_root.extend(
                        [addition_wf_fw_id_mapping[fw_id] for fw_id in addition_fws_root])
                else:  # isinstance(addition_fws_root, int)
                    mapped_addition_fws_root.append(addition_wf_fw_id_mapping[addition_fws_root])

                if addition_fws_root:
                    logger.debug("Mapped addition_wf root fw_ids {} onto newly created fw_ids {}".format(
                        addition_fws_root, mapped_addition_fws_root[-len(addition_fws_root):]))

                self.write_files_prev(addition_wf, fw_spec, root_fw_ids=mapped_addition_fws_root)
                logger.debug("addition_wf:")
                _log_nested_dict(logger.debug, addition_wf.as_dict())

                if apply_mod_spec_to_addition_wf:
                    apply_mod_spec(addition_wf, fw_action, fw_ids=mapped_addition_fws_root)

                addition_wf_list.append(addition_wf)
                all_mapped_addition_fws_root.append(mapped_addition_fws_root)

        # if an action's detour is not None but empty list, then fireworks
        # throws
        #   File ".../site-packages/fireworks/core/firework.py", line 185, in __init__
        #     elif not isinstance(self.detours_root_fw_ids[0], (list, tuple)):
        #       IndexError: list index out of range"
        if len(detour_wf_list) > 0:
            for index, detour_wf in enumerate(detour_wf_list):
                logger.debug("detour_wf {}:".format(index))
            _log_nested_dict(logger.debug, detour_wf.as_dict())

            fw_action.detours = detour_wf_list
            fw_action.detours_root_fw_ids = all_mapped_detour_fws_root
            fw_action.detours_leaf_fw_ids = all_mapped_detour_fws_leaf

        if len(addition_wf_list) > 0:
            for index, addition_wf in enumerate(addition_wf_list):
                logger.debug("addition_wf {}:".format(index))
            _log_nested_dict(logger.debug, addition_wf.as_dict())

            fw_action.additions = addition_wf_list
            fw_action.additions_root_fw_ids = all_mapped_addition_fws_root

        return fw_action


class EnhancedJoinDictTask(FireTaskBase):
    """ combines specified spec fields into a dictionary """
    _fw_name = 'EnhancedJoinDictTask'
    required_params = ['inputs', 'output']
    optional_params = ['rename']

    def run_task(self, fw_spec):
        assert isinstance(self['output'], str)
        assert isinstance(self['inputs'], list)

        try:  # replace if / esle with try / except to find possibly nested val
            output = get_nested_dict_value(fw_spec, self['output'])
        except KeyError:
            output = {}

        assert isinstance(output, dict), "output must be dict."

        if self.get('rename'):
            assert isinstance(self.get('rename'), dict)
            rename = self.get('rename')
        else:
            rename = {}
        for item in self['inputs']:
            if item in rename:
                output = set_nested_dict_value(
                    output, self['rename'][item],
                    get_nested_dict_value(fw_spec, item))
                # replaces
                # output[self['rename'][item]] = fw_spec[item]
            else:
                output = set_nested_dict_value(
                    output, item,
                    get_nested_dict_value(fw_spec, item))
                # replaces
                # output[item] = fw_spec[item]

        return FWAction(mod_spec=[{'_set': {self['output']: output}}])
        # replaces
        # return FWAction(update_spec={self['output']: output})


class EnhancedJoinListTask(FireTaskBase):
    """ combines specified spec fields into a list. """
    _fw_name = 'EnhancedJoinListTask'
    required_params = ['inputs', 'output']

    def run_task(self, fw_spec):
        assert isinstance(self['output'], str)
        assert isinstance(self['inputs'], list)

        try:  # replace if / esle with try / except to find possibly nested val
            output = get_nested_dict_value(fw_spec, self['output'])
        except KeyError:
            output = []
        assert isinstance(output, list), "output must be list."
        # replaces
        # if self['output'] not in fw_spec:
        #    output = []
        # else:
        #    assert isinstance(fw_spec[self['output']], list)
        #    output = fw_spec[self['output']]

        for item in self['inputs']:
            output.append(get_nested_dict_value(fw_spec, item))
            # replaces
            # output.append(fw_spec[item])

        return FWAction(mod_spec=[{'_set': {self['output']: output}}])
        # replaces
        # return FWAction(update_spec={self['output']: output})


# coding: utf-8
#
# multiprocessing.py
#
# Copyright (C) 2020 IMTEK Simulation
# Author: Johannes Hoermann, johannes.hoermann@imtek.uni-freiburg.de
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Multiprocessing utils."""

from abc import abstractmethod
from typing import List


import multiprocessing  # run task as child process to avoid side effects
import traceback  # forward exception from child process to parent process

from fireworks.core.firework import FiretaskBase, FWAction


# in order to make sure any modifications to the environment within task
# won't pollute Fireworks process environmnet, run task in separate process
# https://stackoverflow.com/questions/19924104/python-multiprocessing-handling-child-errors-in-parent
class Process(multiprocessing.Process):
    """
    Class which returns child Exceptions to Parent.
    https://stackoverflow.com/a/33599967/4992248
    """

    def __init__(self, *args, **kwargs):
        multiprocessing.Process.__init__(self, *args, **kwargs)
        self._parent_conn, self._child_conn = multiprocessing.Pipe()
        self._exception = None

    def run(self):
        try:
            multiprocessing.Process.run(self)
            self._child_conn.send(None)
        except Exception as e:
            tb = traceback.format_exc()
            self._child_conn.send((e, tb))
            raise e  # You can still rise this exception if you need to

    @property
    def exception(self):
        if self._parent_conn.poll():
            self._exception = self._parent_conn.recv()
        return self._exception


class RunAsChildProcessTask(FiretaskBase):
    """Run task as child process to avoid side effects by environment pollution.

    Required params:
        None
    Optional params:
        None
    """
    _fw_name = 'RunAsChildProcessTask'
    required_params: List[str] = []
    optional_params: List[str] = []

    @abstractmethod
    def _run_task_as_child_process(
            self,
            fw_spec: dict,
            q: multiprocessing.Queue,
            e: multiprocessing.Event = None
        ) -> None:
        """Replaces run_task in derivatives.

        Child must use q: Queue to return FWAction, that means

            q.put(fw_action)

        replaces

            return fw_action

        If child does not queue fw_action, then parent process will dead lock.
        If set via method 'set_stop_event', then this event is passed on to
        the child to enable it to exit gacefully when desired. It must still
        return an FWAction object.
        """
        ...

    def run_task(self, fw_spec: dict) -> FWAction:
        """Spawn child process to assure my environment stays untouched."""
        q = multiprocessing.Queue()
        e = multiprocessing.Event()
        p = Process(
            target=self._run_task_as_child_process, args=(fw_spec, q, e))

        p.start()

        # wait for child to queue fw_action object and
        # check whether child raises exception
        while q.empty():
            # if child raises exception, then it has terminated
            # before queueing any fw_action object
            if p.exception:
                error, p_traceback = p.exception
                raise ChildProcessError(p_traceback)

            # let child process stop
            if hasattr(self, '_stop_event') and hasattr(self._stop_event, 'is_set'):
                if self._stop_event.is_set():
                    e.set()

        # this loop will deadlock for any child that never raises
        # an exception and does not queue anything

        # queue only used for one transfer of
        # return fw_action, should thus never deadlock.
        fw_action = q.get()
        # if we reach this line without the child
        # queueing anything, then process will deadlock.
        p.join()
        return fw_action

    def set_stop_event(self, e):
        self._stop_event = e



"""dtool lookup server fixtures"""
import os
import sys

import pytest


# TODO: dtool-lookup-server testing instance provision
@pytest.fixture(scope="session")
def dtool_lookup_server(request):
    pass  # stub



#!/usr/bin/env python
#
# tracer.py
#
# Copyright (C) 2020 IMTEK Simulation
# Author: Johannes Hoermann, johannes.hoermann@imtek.uni-freiburg.de
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Offers decorator for tracing with hunter module if available."""

import functools
import sys
try:  # allow for simple tracing when hunter tracer is available
    import hunter
    TRACE = True
except Exception:
    TRACE = False


# inspired by https://python-hunter.readthedocs.io/en/latest/cookbook.html#probe-lightweight-tracing
def trace_func(call_printer_stream=sys.stderr, vars_snooper_stream=sys.stderr, module=__name__):
    """Trace function call to custom stream."""
    def decorator_trace_func(func):
        @functools.wraps(func)
        def wrapper_trace_func(*args, **kwargs):
            # try to set up tracer, fails if 'hunter' module not available
            if TRACE:
                with hunter.trace(
                        module=module,
                        stdlib=False,
                        threading_support=False,
                        actions=[
                            hunter.CallPrinter(stream=call_printer_stream),
                            hunter.VarsSnooper(stream=vars_snooper_stream),
                        ]
                    ):
                    return func(*args, **kwargs)
            else:
                return func(*args, **kwargs)
        return wrapper_trace_func
    return decorator_trace_func



#!/usr/bin/env python
#
# serialize.py
#
# Copyright (C) 2020 IMTEK Simulation
# Author: Johannes Hoermann, johannes.hoermann@imtek.uni-freiburg.de
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# from inspect import getmembers, isfunction
import inspect
import logging
import dill

# https://medium.com/@emlynoregan/serialising-all-the-functions-in-python-cd880a63b591
# https://stackoverflow.com/questions/1253528/is-there-an-easy-way-to-pickle-a-python-function-or-otherwise-serialize-its-cod
# https://www.semicolonworld.com/question/59487/how-to-pickle-a-python-function-with-its-dependencies
# https://stackoverflow.com/questions/45616584/serializing-an-object-in-main-with-pickle-or-dill
# https://stackoverflow.com/questions/11866944/how-to-pickle-functions-classes-defined-in-main-python/19428760

# issue with recurse and imports:
# https://github.com/uqfoundation/dill/issues/219


def _fixed_create_function(fcode, fglobals, fname=None, fdefaults=None,
                     fclosure=None, fdict=None, fkwdefaults=None):
    # same as FunctionType, but enable passing __dict__ to new function,
    # __dict__ is the storehouse for attributes added after function creation
    if fdict is None: fdict = dict()
    FunctionType = __import__('types').FunctionType
    func = FunctionType(fcode, fglobals or dict(), fname, fdefaults, fclosure)
    func.__dict__.update(fdict) #XXX: better copy? option to copy?
    if fkwdefaults is not None:
        func.__kwdefaults__ = fkwdefaults

    # THE WORKAROUND:
    # if the function was serialized without recurse, fglobals would actually contain
    # __builtins__, but because of recurse only the referenced modules/objects
    # end up in fglobals and we are missing the important __builtins__
    if "__builtins__" not in func.__globals__:
        func.__globals__["__builtins__"] = globals()["__builtins__"]
    return func


dill._dill._create_function = _fixed_create_function


def get_module_member_list(module):
    """Get list of objects within 'module'."""
    # logger = logging.getLogger(__name__)

    def predicate(another_obj):
        return inspect.getmodule(another_obj) == module

    obj_list = inspect.getmembers(module, predicate)
    return obj_list


def serialize_module_obj(obj):
    """Serialize all objects within module of 'obj' along with 'obj'."""
    logger = logging.getLogger(__name__)
    module = inspect.getmodule(obj)
    logger.info("{:s} belongs to {:s}".format(obj.__name__, module.__name__))

    # list of 'name', 'object' tuples
    member_list = get_module_member_list(module)
    member_dict = dict(member_list)
    logger.info("{} in same module as {:s}".format(
        list(member_dict.keys()), obj.__name__))

    # a hack in order to get dill to serialize other objects within same module
    # as 'obj' as well if 'obj' depends on them.
    # dill looks for dependencies only within __main__
    main_backup = {}
    main_deletes = []

    import __main__

    for name, module_member in member_dict.items():
        if hasattr(__main__, name):
            main_backup[name] = getattr(__main__, name)
        else:
            main_deletes.append(name)

        if hasattr(module_member, '__module__'):
            module_member.__module__ = __main__.__name__
        setattr(__main__, name, module_member)

    func_str = dill.dumps(
        getattr(__main__, obj.__name__),
        protocol=dill.HIGHEST_PROTOCOL, byref=False, recurse=True)

    # reset __module__ attribute:
    for name, module_member in member_dict.items():
        if hasattr(module_member, '__module__'):
            module_member.__module__ = module.__name__

    for name, main_member in main_backup.items():
        setattr(__main__, name, main_member)

    for name in main_deletes:
        delattr(__main__, name)

    return func_str


def serialize_obj(obj):
    """Serialize an object without following any dependencies."""
    func_str = dill.dumps(
        obj, protocol=dill.HIGHEST_PROTOCOL, byref=False, recurse=True)
    return func_str



# coding: utf-8
#
# ssh_tasks.py
#
# Copyright (C) 2020 IMTEK Simulation
# Author: Johannes Hoermann, johannes.hoermann@imtek.uni-freiburg.de
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
__author__ = 'Johannes Laurin Hoermann'
__copyright__ = 'Copyright 2020, IMTEK Simulation, University of Freiburg'
__email__ = 'johannes.hoermann@imtek.uni-freiburg.de, johannes.laurin@gmail.com'
__date__ = 'May 7, 2020'

from contextlib import ExitStack

import io
import logging
import os.path
import select
import socket
import threading
import time

try:
    import SocketServer
except ImportError:
    import socketserver as SocketServer
import paramiko

from fireworks.fw_config import FW_LOGGING_FORMAT
from fireworks.core.firework import FWAction
from fireworks.utilities.dict_mods import get_nested_dict_value
from fireworks.utilities.fw_serializers import ENCODING_PARAMS

from jlhfw.utils.multiprocessing import RunAsChildProcessTask
from jlhfw.utils.logging import LoggingContext

DEFAULT_FORMATTER = logging.Formatter(FW_LOGGING_FORMAT)


def from_fw_spec(param, fw_spec):
    """Expands param['key'] as key within fw_spec.

    If param is dict hand has field 'key', then return value at specified
    position from fw_spec. Otherwise, return 'param' itself.
    """
    if isinstance(param, dict) and 'key' in param:
        ret = get_nested_dict_value(fw_spec, param['key'])
    else:
        ret = param
    return ret


class ForwardServer(SocketServer.ThreadingTCPServer):
    """
    This class has been copied and modified from the original paramiko
    repository at
    https://github.com/paramiko/paramiko/blob/master/demos/forward.py
    and underlies the GNU Lesser General Public Licesnse v 2.1 at
    https://github.com/paramiko/paramiko/blob/master/LICENSE.
    """
    daemon_threads = True
    allow_reuse_address = True


class Handler(SocketServer.BaseRequestHandler):
    """
    This class has been copied and modified from the original paramiko
    repository at
    https://github.com/paramiko/paramiko/blob/master/demos/forward.py
    and underlies the GNU Lesser General Public Licesnse v 2.1 at
    https://github.com/paramiko/paramiko/blob/master/LICENSE.
    """
    def handle(self):
        logger = logging.getLogger(__name__)
        try:
            chan = self.ssh_transport.open_channel(
                "direct-tcpip",
                (self.chain_host, self.chain_port),
                self.request.getpeername(),
            )
        except Exception as e:
            logger.info(
                "Incoming request to %s:%d failed: %s"
                % (self.chain_host, self.chain_port, repr(e))
            )
            return
        if chan is None:
            logger.info(
                "Incoming request to %s:%d was rejected by the SSH server."
                % (self.chain_host, self.chain_port)
            )
            return

        logger.info(
            "Connected!  Tunnel open %r -> %r -> %r"
            % (
                self.request.getpeername(),
                chan.getpeername(),
                (self.chain_host, self.chain_port),
            )
        )
        while True:
            r, w, x = select.select([self.request, chan], [], [])
            if self.request in r:
                data = self.request.recv(1024)
                if len(data) == 0:
                    break
                chan.send(data)
            if chan in r:
                data = chan.recv(1024)
                if len(data) == 0:
                    break
                self.request.send(data)

        peername = self.request.getpeername()
        chan.close()
        self.request.close()
        logger.info("Tunnel closed from %r" % (peername,))


class SSHForwardTask(RunAsChildProcessTask):
    """
    A Firetask to open ssh port forwarding via jump host.

    Required params:
        - remote_host: target remote host
        - remote_port: target port at remote host
        - ssh_host: intermediate ssh jump host
        - ssh_user: ssh user at intermediate jump host
    Optional params:
        - local_port (int, default: None): local port, allocate random free
            port if not specified.
        - port_file (str, default: '.port'): write allocated local
            port number to that file.
        - ssh_port (int, default: 22): ssh port at intermediate jump host
        - ssh_keyfile (st, default '~/.ssh/id_rsa'): ssh identity file to
            establish connection to intermediate jump host

        All of the above required and optional parameters may also be dicts of
        format { 'key': 'some->nested->fw_spec->key' } for looking up values
        within fw_spec instead.

        - store_stdlog (bool, default: False): insert log output into database
        - stdlog_file (str, Default: NameOfTaskClass.log): print log to file
        - loglevel (str, Default: logging.INFO): loglevel for this task

    Example:

        t = SSHForwardTask(
            remote_host="ufr2.isi1.public.ads.uni-freiburg.de",
            remote_port=445,
            local_port=10445,
            ssh_host="simdata.vm.uni-freiburg.de",
            ssh_user="sshclient",
            ssh_keyfile="~/.ssh/sshclient-frrzvm",
            ssh_port=22,
            port_file='.port')

        establishes the Paramiko equivalent of the command line ssh call

            ssh  -i ~/.ssh/sshclient-frrzvm -N -L \
              10445:ufr2.isi1.public.ads.uni-freiburg.de:445 \
              sshclient@132.230.102.164

        Set 'local_port' to None to allocate a random free port. Read that
        port from 'port_file' afte connection has been established.
    """
    required_params = [
        *RunAsChildProcessTask.required_params,
        'remote_host',
        'remote_port',
        'ssh_host',
        'ssh_user',
    ]

    optional_params = [
        *RunAsChildProcessTask.optional_params,
        'local_port',
        'port_file',
        'ssh_port',
        'ssh_keyfile',

        'store_stdlog',
        'stdlog_file',
        'loglevel',
    ]

    def _run_task_as_child_process(self, fw_spec, q, e=None):
        """q is a Queue used to return fw_action, e is an Event to end task."""
        # required
        remote_host = from_fw_spec(self.get('remote_host'), fw_spec)
        remote_port = from_fw_spec(self.get('remote_port'), fw_spec)
        ssh_host = from_fw_spec(self.get('ssh_host'), fw_spec)
        ssh_user = from_fw_spec(self.get('ssh_user'), fw_spec)

        # optional with defaults
        local_port = from_fw_spec(self.get("local_port", None), fw_spec)
        port_file = from_fw_spec(self.get("port_file", '.port'), fw_spec)

        ssh_port = from_fw_spec(self.get("ssh_port", 22), fw_spec)
        ssh_port = int(ssh_port)
        assert isinstance(ssh_port, int)

        ssh_keyfile = from_fw_spec(
            self.get("ssh_keyfile", '~/.ssh/id_rsa'), fw_spec)

        stdlog_file = self.get('stdlog_file', '{}.log'.format(self._fw_name))
        store_stdlog = self.get('store_stdlog', False)

        loglevel = self.get('loglevel', logging.INFO)

        with ExitStack() as stack:

            if store_stdlog:
                stdlog_stream = io.StringIO()
                logh = logging.StreamHandler(stdlog_stream)
                logh.setFormatter(DEFAULT_FORMATTER)
                stack.enter_context(
                    LoggingContext(handler=logh, level=loglevel, close=False))

            # logging to dedicated log file if desired
            if stdlog_file:
                logfh = logging.FileHandler(stdlog_file, mode='a', **ENCODING_PARAMS)
                logfh.setFormatter(DEFAULT_FORMATTER)
                stack.enter_context(
                    LoggingContext(handler=logfh, level=loglevel, close=True))

            logger = logging.getLogger(__name__)

            # allocate free local port if none specified
            # inspired by
            #   https://www.scivision.dev/python-get-free-network-port
            # and
            #   http://code.activestate.com/recipes/531822-pick-unused-port
            if local_port is None:
                s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                s.bind(('localhost', 0))
                s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                addr, local_port = s.getsockname()
                logger.info(
                    "Allocated free local port: {:d}".format(local_port))

            # write port number to file for communication with other processes
            if isinstance(port_file, str):
                with open(port_file, mode='w') as f:
                    f.write(str(local_port))
                logger.info("Wrote port number to '{}'".format(port_file))

            transport = paramiko.Transport((ssh_host, ssh_port))
            pkey = paramiko.RSAKey.from_private_key_file(
                os.path.expanduser(ssh_keyfile))

            transport.connect(hostkey=None,
                              username=ssh_user,
                              password=None,
                              pkey=pkey)

            class SubHander(Handler):
                chain_host = remote_host
                chain_port = remote_port
                ssh_transport = transport

            with ForwardServer(("", local_port), SubHander) as server:
                # if a stop signal is to be expected, then run server in
                # separate thread and forward stop signal
                if e is not None:
                    server_thread = threading.Thread(
                        target=server.serve_forever)
                    # Exit the server thread when the main thread terminates
                    server_thread.daemon = True
                    server_thread.start()
                    logger.info(
                        "Server loop running in thread '{}'".format(
                            server_thread.name))

                    while server_thread.is_alive():
                        if e.is_set():
                            logger.info("Send stop signal to server loop.")
                            server.shutdown()
                        time.sleep(1)
                else:
                    server.serve_forever()

                logger.info("Server stoped.")

            try:
                logger.info("Trying to close socket.")
                s.close()  # fails if s does not exist
            except Exception:
                logger.info("Apparently already closed.")

        q.put(FWAction())



# coding: utf-8
#
# dtool_lookup_tasks.py
#
# Copyright (C) 2020 IMTEK Simulation
# Author: Johannes Hoermann, johannes.hoermann@imtek.uni-freiburg.de
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Simple dtool-lookup-client interface for querying a dtool lookup server."""
from typing import Dict, List

from abc import abstractmethod
from contextlib import ExitStack

import io
import json
import logging
import os
import pymongo

from ruamel.yaml import YAML

from fireworks.fw_config import FW_LOGGING_FORMAT

from fireworks.core.firework import FWAction
from fireworks.utilities.dict_mods import get_nested_dict_value
from fireworks.utilities.fw_serializers import ENCODING_PARAMS

from jlhfw.utils.multiprocessing import RunAsChildProcessTask
from jlhfw.utils.dict import dict_merge
from jlhfw.utils.environ import TemporaryOSEnviron
from jlhfw.utils.logging import LoggingContext, _log_nested_dict


__author__ = 'Johannes Laurin Hoermann'
__copyright__ = 'Copyright 2020, IMTEK Simulation, University of Freiburg'
__email__ = 'johannes.hoermann@imtek.uni-freiburg.de, johannes.laurin@gmail.com'
__date__ = 'Nov 03, 2020'


# TODO: dates might be string or float, fix on server side
Date = (str, float)

DEFAULT_FORMATTER = logging.Formatter(FW_LOGGING_FORMAT)

# fields expected in dtool-lookup-server's response
FIELD_TYPE_DICT = {
    'base_uri': str,
    'created_at': Date,
    'creator_username': str,
    'frozen_at': Date,
    'name': str,
    'uri': str,
    'uuid': str,
}

yaml = YAML()


def write_serialized(obj, file, format=None):
    """Write serializable object to file. Format from file suffix."""
    if not format:
        _, extension = os.path.splitext(file)
        format = extension[1:].strip().lower()

    with open(file, "w") as f:
        if format in ['json']:
            json.dump(obj, f)
        else:
            yaml.dump(obj, f)


def from_fw_spec(param, fw_spec):
    """Expands param['key'] as key within fw_spec.

    If param is dict hand has field 'key', then return value at specified
    position from fw_spec. Otherwise, return 'param' itself.
    """
    if isinstance(param, dict) and 'key' in param:
        ret = get_nested_dict_value(fw_spec, param['key'])
    else:
        ret = param
    return ret


# TODO: this should rather go into dtool-lookup-client itself
def validate_dataset_info(d):
    """Check all expected fields exist."""
    for f, t in FIELD_TYPE_DICT.items():
        if f not in d:
            raise KeyError("Expected key '{}' not in '{}'".format(f, d))

        if not isinstance(d[f], t):
            raise KeyError(
                "Key '{}' value type('{}') is '{}', but '{}' expected.".format(
                    f, d[f], type(d[f]), t))


# TODO: This follows almost the same pattern as DtoolTask, further abstraction possible
class DtoolLookupTask(RunAsChildProcessTask):
    """
    A dtool lookup task ABC.

    Required params:
        None
    Optional params:
        - dtool_config (dict): dtool config key-value pairs, override
            defaults in $HOME/.config/dtool/dtool.json. Default: None
        - dtool_config_key (str): key to dict within fw_spec, override
            defaults in $HOME/.config/dtool/dtool.json and static dtool_config
            task spec. Default: None.
        - output (str): spec key that will be used to pass
            the query result's content to child fireworks. Default: None
        - dict_mod (str, default: '_set'): how to insert handled dataset's
            properties into output key, see fireworks.utils.dict_mods
        - propagate (bool, default:None): if True, then set the
            FWAction 'propagate' flag and propagate updated fw_spec not only to
            direct children, but to all descendants down to wokflow's leaves.
        - stored_data (bool, default: False): put handled dataset properties
            into FWAction.stored_data
        - store_stdlog (bool, default: False): insert log output into FWAction.stored_data
        - stdlog_file (str, Default: NameOfTaskClass.log): print log to file
        - loglevel (str, Default: logging.INFO): loglevel for this task
    """
    _fw_name = 'DtoolLookupTask'
    required_params = [*RunAsChildProcessTask.required_params]
    optional_params = [
        *RunAsChildProcessTask.optional_params,
        "dtool_config",
        "dtool_config_key",
        "stored_data",
        "output",
        "dict_mod",
        "propagate",
        "stdlog_file",
        "store_stdlog",
        "loglevel"]

    @abstractmethod
    def _run_task_internal(self, fw_spec) -> List[Dict]:
        """Derivatives implement their functionality here."""
        ...

    def _run_task_as_child_process(self, fw_spec, q, e=None):
        """q is a Queue used to return fw_action."""
        stored_data = self.get('stored_data', False)
        output_key = self.get('output', None)
        dict_mod = self.get('dict_mod', '_set')
        propagate = self.get('propagate', False)

        dtool_config = self.get("dtool_config", {})
        dtool_config_key = self.get("dtool_config_key")

        stdlog_file = self.get('stdlog_file', '{}.log'.format(self._fw_name))
        store_stdlog = self.get('store_stdlog', False)

        loglevel = self.get('loglevel', logging.INFO)

        with ExitStack() as stack:

            if store_stdlog:
                stdlog_stream = io.StringIO()
                logh = logging.StreamHandler(stdlog_stream)
                logh.setFormatter(DEFAULT_FORMATTER)
                stack.enter_context(
                    LoggingContext(handler=logh, level=loglevel, close=False))

            # logging to dedicated log file if desired
            if stdlog_file:
                logfh = logging.FileHandler(stdlog_file, mode='a', **ENCODING_PARAMS)
                logfh.setFormatter(DEFAULT_FORMATTER)
                stack.enter_context(
                    LoggingContext(handler=logfh, level=loglevel, close=True))

            logger = logging.getLogger(__name__)

            logger.debug("task spec level dtool config overrides:")
            _log_nested_dict(logger.debug, dtool_config)

            # fw_spec dynamic dtool_config overrides
            dtool_config_update = {}
            if dtool_config_key is not None:
                try:
                    dtool_config_update = get_nested_dict_value(
                        fw_spec, dtool_config_key)
                    logger.debug("fw_spec level dtool config overrides:")
                    _log_nested_dict(logger.debug, dtool_config_update)
                except Exception:  # key not found
                    logger.warning("{} not found within fw_spec, ignored.".format(
                        dtool_config_key))
            dtool_config.update(dtool_config_update)
            logger.debug("effective dtool config overrides:")
            _log_nested_dict(logger.debug, dtool_config)

            stack.enter_context(TemporaryOSEnviron(env=dtool_config))

            output = self._run_task_internal(fw_spec)

        fw_action = FWAction()

        if stored_data or store_stdlog:
            fw_action.stored_data = {}
        if store_stdlog:
            stdlog_stream.flush()
            fw_action.stored_data['stdlog'] = stdlog_stream.getvalue()
        if stored_data:
            fw_action.stored_data['output'] = output

        # 'propagate' only development feature for now
        if hasattr(fw_action, 'propagate') and propagate:
            fw_action.propagate = propagate

        if output_key:  # inject into fw_spec
            fw_action.mod_spec = [{dict_mod: {output_key: output}}]

        # return fw_action
        q.put(fw_action)


class QueryDtoolTask(DtoolLookupTask):
    """
    A Firetask to query datasets from via dtool lookup server and write the
    them to specified directory (current working directory if not specified).

    Required params:

    Optional params:
        - query (dict): mongo db query identifying files to fetch.
          Same as within fireworks.utils.dict_mods, use '->' in dict keys
          for querying nested documents, instead of MongoDB '.' (dot) seperator.
          Do use '.' and NOT '->' within the 'sort_key' field.
        - query_key (str): entry in fw_spec this key points to supersedes 'query' if specified
        - sort_key (str): sort key, sort by 'frozen_at' per default
        - sort_direction (int): sort direction, default 'pymongo.DESCENDING'
        - limit (int): maximum number of files to write, default: no limit
        - fizzle_empty_result (bool): fizzle if nothing found, default: True
        - fizzle_degenerate_dataset_name (bool): fizzle if more than one of the
          resulting datasets are named equivalently, default: False
        - metadata_file (str): default: None. If specififed, then metadata of
          queried files is written in .yaml or .json format, depending on the
          file name suffix.
        - expand (bool): will replace list result with single value result if
          list result only contains one entry and with None if list result is empty.
          Default: False.

    Fields 'sort_key', 'sort_direction', 'limit', 'fizzle_empty_result',
    'fizzle_degenerate_dataset_name', 'meta_file', 'expand', may also be a dict of format
    { 'key': 'some->nested->fw_spec->key' } for looking up value within
    fw_spec instead.
    """

    _fw_name = 'QueryDtoolTask'
    required_params = [
        *DtoolLookupTask.required_params,
    ]
    optional_params = [
        *DtoolLookupTask.optional_params,
        "query", "query_key",
        "fizzle_degenerate_dataset_name", "fizzle_empty_result",
        "limit", "metadata_file",
        "sort_direction", "sort_key", "expand"]

    def _run_task_internal(self, fw_spec):
        import dtool_lookup_api
        dtool_lookup_api.core.config.Config.interactive = False

        logger = logging.getLogger(__name__)

        query = self.get("query", {})
        query_key = self.get("query_key", None)

        sort_key = self.get("sort_key", 'frozen_at')
        sort_key = from_fw_spec(sort_key, fw_spec)
        if sort_key not in FIELD_TYPE_DICT:
            raise ValueError(
                "'sort_key' is '{}', but must be selected from '{}'".format(
                    sort_key, list(FIELD_TYPE_DICT.keys())
                ))

        sort_direction = self.get("sort_direction", pymongo.DESCENDING)
        sort_direction = from_fw_spec(sort_direction, fw_spec)
        if sort_direction not in [pymongo.ASCENDING, pymongo.DESCENDING]:
            raise ValueError(
                "'sort_direction' is '{}', but must be selected from '{}'".format(
                    sort_key, [pymongo.ASCENDING, pymongo.DESCENDING]
                ))

        limit = self.get("limit", None)
        limit = from_fw_spec(limit, fw_spec)

        fizzle_empty_result = self.get("fizzle_empty_result", True)
        fizzle_empty_result = from_fw_spec(fizzle_empty_result, fw_spec)

        fizzle_degenerate_dataset_name = self.get("fizzle_degenerate_dataset_name", True)

        meta_file = self.get("meta_file", None)
        meta_file = from_fw_spec(meta_file, fw_spec)

        expand = self.get('expand', None)
        expand = from_fw_spec(expand, fw_spec)

        if query_key:
            logger.debug("query from fw_spec at '%s'." % query_key)
            query = get_nested_dict_value(fw_spec, query_key)
        elif input:
            pass
        else:
            raise ValueError("Neither 'query' nor 'query_key' specified.")

        r = dtool_lookup_api.query(query)

        logger.debug("Server response: '{}'".format(r))

        if fizzle_empty_result and (len(r) == 0):
            raise ValueError("Query yielded empty result! (query: {})".format(
                query))

        for doc in r:
            logger.debug("Validating '{}'...".format(doc))
            validate_dataset_info(doc)

        if sort_key:
            logger.debug("Sorting by key '{}' and direction '{}'.".format(
                         sort_key, sort_direction))
            r = sorted(r, key=lambda d: d[sort_key],
                       reverse=(sort_direction == pymongo.DESCENDING))
            logger.debug("Sorted server response: '{}'".format(r))

        if limit:
            logger.debug("Limiting response to the first {} enrties...".format(
                         limit))
            r = r[:limit]
            logger.debug("Truncated server response: '{}'".format(r))

        unique_dataset_names = set()  # track all used file names

        for i, doc in enumerate(r):
            dataset_name = doc["name"]
            if fizzle_degenerate_dataset_name and (dataset_name in unique_dataset_names):
                raise ValueError((
                    "The dataset name {} is used "
                    "a second time by result {:d}/{:d}! (query: {})").format(
                        dataset_name, i, len(r), query))

            unique_dataset_names.add(dataset_name)

        if expand and len(r) == 1:
            r = r[0]
            logger.debug("Expand single-entry result'%s'." % r)
        elif expand and len(r) == 0:
            r = None
            logger.debug("Expand empty result as None.")

        if meta_file:
            logger.debug("Write response to file '{}'.".format(meta_file))
            write_serialized(r, meta_file)

        return r


class ReadmeDtoolTask(DtoolLookupTask):
    """
    Firetask to query a dataset's README via dtool lookup server and inject
    content into fw_spec.

    Required params:
        - uri (str): dataset URI

    Optional params:
        - dest_dir (str): destination directory, default is the current working
            directory.
        - fizzle_empty_result (bool): fizzle if nothing found, default: True
        - fw_supersedes_dtool: if True, metadata from metadata_fw_source_key
            supersedes metadata from metadata_fp_source_key. Otherwise, latter
            supersedes former. Default: False
        - metadata_dtool_exclusions (dict): mark exclusions from metadata forwarded
            into pipeline with nested dict {'excluded': {'field': True}}.
        - metadata_dtool_source_key (str): if specified, only use nested subset of
            metadata found in README. Default: None or "" uses all.
        - metadata_file (str): default: None. If specififed, then metadata of
            queried files is written in .yaml or .json format, depending on the
            file name suffix.
        - metadata_fw_exclusions (dict): mark exclusions from
            metadata_fw_source_key merged with metadata_fp_source_key.
        - metadata_fw_source_key (str): merge metadata from README with
            metadata within this fw_spec key before forwarding.

    Fields 'uri', 'sort_direction', 'dest_dir', 'fizzle_empty_result',
    'fw_supersedes_dtool', 'metadata_dtool_source_key', 'metadata_file',
    'metadata_fw_source_key' may also be a dict of format
    { 'key': 'some->nested->fw_spec->key' } for looking up value within
    fw_spec instead.
    """
    _fw_name = 'ReadmeDtoolTask'
    required_params = [
        *DtoolLookupTask.required_params,
        "uri"]
    optional_params = [
        *DtoolLookupTask.optional_params,
        "fizzle_empty_result",
        "metadata_file",

        "metadata_dtool_source_key",
        "metadata_fw_source_key",
        "fw_supersedes_dtool",
        "metadata_dtool_exclusions",
        "metadata_fw_exclusions",
    ]

    def _run_task_internal(self, fw_spec):
        import dtool_lookup_api
        dtool_lookup_api.core.config.Config.interactive = False

        logger = logging.getLogger(__name__)

        uri = self.get("uri", None)
        uri = from_fw_spec(uri, fw_spec)

        dest_dir = self.get("dest_dir", os.path.abspath(os.path.curdir))
        dest_dir = from_fw_spec(dest_dir, fw_spec)

        fizzle_empty_result = self.get("fizzle_empty_result", True)
        fizzle_empty_result = from_fw_spec(fizzle_empty_result, fw_spec)

        metadata_file = self.get("metadata_file", None)
        metadata_file = from_fw_spec(metadata_file, fw_spec)

        metadata_dtool_source_key = self.get("metadata_dtool_source_key", None)
        metadata_dtool_source_key = from_fw_spec(metadata_dtool_source_key, fw_spec)

        metadata_fw_source_key = self.get("metadata_fw_source_key", None)
        metadata_fw_source_key = from_fw_spec(metadata_fw_source_key, fw_spec)

        fw_supersedes_dtool = self.get("fw_supersedes_dtool", False)
        fw_supersedes_dtool = from_fw_spec(fw_supersedes_dtool, fw_spec)

        metadata_dtool_exclusions = self.get("metadata_dtool_exclusions", {})
        metadata_fw_exclusions = self.get("metadata_fw_exclusions", {})

        readme = dtool_lookup_api.readme(uri)
        metadata = {}
        if isinstance(readme, str) and fizzle_empty_result:  # server error message
            raise ValueError(readme)
        elif isinstance(readme, str):
            logger.warning(readme)
        else:
            logger.debug("Metadata of queried object:")
            _log_nested_dict(logger.debug, readme)

            if metadata_file:
                metadata_file_path = os.path.join(dest_dir, metadata_file)
                logger.debug("Write metadata of queried dataset to file '{}'.".format(metadata_file_path))
                write_serialized(readme, metadata_file_path)

            if metadata_dtool_source_key and metadata_dtool_source_key != "":
                dtool_metadata = get_nested_dict_value(readme, metadata_dtool_source_key)
                logger.debug(
                    "Forwarding '{}' subset of README metadata:".format(metadata_dtool_source_key))
                _log_nested_dict(logger.debug, dtool_metadata)
            else:
                dtool_metadata = readme
                logger.debug("Forwarding README metadata.")

            if metadata_fw_source_key and metadata_fw_source_key == "":
                fw_metadata = fw_spec
                logger.debug("Merging with fw_spec.")
            elif metadata_fw_source_key:
                fw_metadata = get_nested_dict_value(fw_spec, metadata_fw_source_key)
                logger.debug("Merging with '{}' subset of fw_spec:".format(metadata_fw_source_key))
                _log_nested_dict(logger.debug, fw_metadata)
            else:
                fw_metadata = {}

            if fw_supersedes_dtool:
                metadata = dict_merge(dtool_metadata, fw_metadata,
                                      exclusions=metadata_dtool_exclusions,
                                      merge_exclusions=metadata_fw_exclusions)
            else:
                metadata = dict_merge(fw_metadata, dtool_metadata,
                                      exclusions=metadata_fw_exclusions,
                                      merge_exclusions=metadata_dtool_exclusions)

            logger.debug("Forwarding merge:")
            _log_nested_dict(logger.debug, metadata)

        return metadata


class ManifestDtoolTask(DtoolLookupTask):
    """
    Firetask to query a dataset's manifest via dtool lookup server. A manifest may look like this:

        {
            'dtoolcore_version': '3.17.0',
            'hash_function': 'md5sum_hexdigest',
            'items': {
                '0a03867efe1eb36d257d4cecf598019963a1f14d': {
                    'hash': '442170a2fd3fa9a54a711ffa1b249e43',
                    'relpath': 'surfactant_head_surfactant_head_rdf.txt',
                    'size_in_bytes': 115098,
                    'utc_timestamp': 1600729024.880694},
                '0b6c156fb7e71a1387904cc6d918cb9acf286538': {
                    'hash': 'bf9cb747ae826511958790034e62d773',
                    'relpath': 'counterion_counterion_rdf.txt',
                    'size_in_bytes': 115098,
                    'utc_timestamp': 1600729024.979725
                },
                ...
            }
        }

    Required params:
        - uri (str): dataset URI

    Optional params:
        - dest_dir (str): destination directory, default is the current working
            directory.
        - fizzle_empty_result (bool): fizzle if nothing found, default: True
        - manifest_file (str): default: None. If specififed, then manifest of
            dataset is written in .yaml or .json format, depending on the
            file name suffix.

        Fields 'uri', 'dest_dir', 'manifest_file',
        may also be a dict of format { 'key': 'some->nested->fw_spec->key' }
        for looking up value within fw_spec instead.
    """
    _fw_name = 'ManifestDtoolTask'
    required_params = [
        *DtoolLookupTask.required_params,
        "uri"]
    optional_params = [
        *DtoolLookupTask.optional_params,
        "dest_dir",
        "fizzle_empty_result",
        "manifest_file",
    ]

    def _run_task_internal(self, fw_spec):
        import dtool_lookup_api
        dtool_lookup_api.core.config.Config.interactive = False

        logger = logging.getLogger(__name__)

        uri = self.get("uri", None)
        uri = from_fw_spec(uri, fw_spec)

        dest_dir = self.get("dest_dir", os.path.abspath(os.path.curdir))
        dest_dir = from_fw_spec(dest_dir, fw_spec)

        fizzle_empty_result = self.get("fizzle_empty_result", True)
        fizzle_empty_result = from_fw_spec(fizzle_empty_result, fw_spec)

        manifest_file = self.get("manifest_file", None)
        manifest_file = from_fw_spec(manifest_file, fw_spec)

        manifest = dtool_lookup_api.manifest(uri)
        if isinstance(manifest, str) and fizzle_empty_result:  # server error message
            raise ValueError(manifest)
        elif isinstance(manifest, str):
            logger.warning(manifest)
            manifest = None
        else:  # success
            logger.debug("Manifest of queried dataset:")
            _log_nested_dict(logger.debug, manifest)

            if manifest_file:
                manifest_file_path = os.path.join(dest_dir, manifest_file)
                logger.debug("Writing manifest of queried dataset to '{}'.".format(manifest_file_path))
                write_serialized(manifest, manifest_file_path)

        return manifest


# below tasks are equivalent to above, but do not require a dtool lookup server connection.
# Instead, they generate the desired metadata directly from a reachable dataset.

# TODO: factor out common parts of direct tasks below and lookup tasks above (almost everything)

class DirectReadmeTask(DtoolLookupTask):
    """
    Firetask to get a dataset's README directly by a dataset's URI and inject
    content into fw_spec.

    Required params:
        - uri (str): dataset URI

    Optional params:
        - dest_dir (str): destination directory, default is the current working
            directory.
        - fizzle_empty_result (bool): fizzle if nothing found, default: True
        - fw_supersedes_dtool: if True, metadata from metadata_fw_source_key
            supersedes metadata from metadata_fp_source_key. Otherwise, latter
            supersedes former. Default: False
        - metadata_dtool_exclusions (dict): mark exclusions from metadata forwarded
            into pipeline with nested dict {'excluded': {'field': True}}.
        - metadata_dtool_source_key (str): if specified, only use nested subset of
            metadata found in README. Default: None or "" uses all.
        - metadata_file (str): default: None. If specififed, then metadata of
            queried files is written in .yaml or .json format, depending on the
            file name suffix.
        - metadata_fw_exclusions (dict): mark exclusions from
            metadata_fw_source_key merged with metadata_fp_source_key.
        - metadata_fw_source_key (str): merge metadata from README with
            metadata within this fw_spec key before forwarding.

    Fields 'uri', 'sort_direction', 'dest_dir', 'fizzle_empty_result',
    'fw_supersedes_dtool', 'metadata_dtool_source_key', 'metadata_file',
    'metadata_fw_source_key' may also be a dict of format
    { 'key': 'some->nested->fw_spec->key' } for looking up value within
    fw_spec instead.
    """
    _fw_name = 'DirectReadmeTask'
    required_params = [
        *DtoolLookupTask.required_params,
        "uri"]
    optional_params = [
        *DtoolLookupTask.optional_params,
        "fizzle_empty_result",
        "metadata_file",

        "metadata_dtool_source_key",
        "metadata_fw_source_key",
        "fw_supersedes_dtool",
        "metadata_dtool_exclusions",
        "metadata_fw_exclusions",
    ]

    def _run_task_internal(self, fw_spec):
        import dtoolcore
        logger = logging.getLogger(__name__)

        uri = self.get("uri", None)
        uri = from_fw_spec(uri, fw_spec)

        dest_dir = self.get("dest_dir", os.path.abspath(os.path.curdir))
        dest_dir = from_fw_spec(dest_dir, fw_spec)

        fizzle_empty_result = self.get("fizzle_empty_result", True)
        fizzle_empty_result = from_fw_spec(fizzle_empty_result, fw_spec)

        metadata_file = self.get("metadata_file", None)
        metadata_file = from_fw_spec(metadata_file, fw_spec)

        metadata_dtool_source_key = self.get("metadata_dtool_source_key", None)
        metadata_dtool_source_key = from_fw_spec(metadata_dtool_source_key, fw_spec)

        metadata_fw_source_key = self.get("metadata_fw_source_key", None)
        metadata_fw_source_key = from_fw_spec(metadata_fw_source_key, fw_spec)

        fw_supersedes_dtool = self.get("fw_supersedes_dtool", False)
        fw_supersedes_dtool = from_fw_spec(fw_supersedes_dtool, fw_spec)

        metadata_dtool_exclusions = self.get("metadata_dtool_exclusions", {})
        metadata_fw_exclusions = self.get("metadata_fw_exclusions", {})

        src_dataset = dtoolcore.DataSet.from_uri(uri)
        readme = yaml.load(src_dataset.get_readme_content())
        metadata = {}
        if isinstance(readme, str) and fizzle_empty_result:  # server error message
            raise ValueError(readme)
        elif isinstance(readme, str):
            logger.warning(readme)
        else:
            logger.debug("Metadata of queried object:")
            _log_nested_dict(logger.debug, readme)

            if metadata_file:
                metadata_file_path = os.path.join(dest_dir, metadata_file)
                logger.debug("Write metadata of queried dataset to file '{}'.".format(metadata_file_path))
                write_serialized(readme, metadata_file_path)

            if metadata_dtool_source_key and metadata_dtool_source_key != "":
                dtool_metadata = get_nested_dict_value(readme, metadata_dtool_source_key)
                logger.debug(
                    "Forwarding '{}' subset of README metadata:".format(metadata_dtool_source_key))
                _log_nested_dict(logger.debug, dtool_metadata)
            else:
                dtool_metadata = readme
                logger.debug("Forwarding README metadata.")

            if metadata_fw_source_key and metadata_fw_source_key == "":
                fw_metadata = fw_spec
                logger.debug("Merging with fw_spec.")
            elif metadata_fw_source_key:
                fw_metadata = get_nested_dict_value(fw_spec, metadata_fw_source_key)
                logger.debug("Merging with '{}' subset of fw_spec:".format(metadata_fw_source_key))
                _log_nested_dict(logger.debug, fw_metadata)
            else:
                fw_metadata = {}

            if fw_supersedes_dtool:
                metadata = dict_merge(dtool_metadata, fw_metadata,
                                      exclusions=metadata_dtool_exclusions,
                                      merge_exclusions=metadata_fw_exclusions)
            else:
                metadata = dict_merge(fw_metadata, dtool_metadata,
                                      exclusions=metadata_fw_exclusions,
                                      merge_exclusions=metadata_dtool_exclusions)

            logger.debug("Forwarding merge:")
            _log_nested_dict(logger.debug, metadata)

        return metadata


class DirectManifestTask(DtoolLookupTask):
    """
    Firetask to generate a dataset's manifest directly. A manifest may look like this:

        {
            'dtoolcore_version': '3.17.0',
            'hash_function': 'md5sum_hexdigest',
            'items': {
                '0a03867efe1eb36d257d4cecf598019963a1f14d': {
                    'hash': '442170a2fd3fa9a54a711ffa1b249e43',
                    'relpath': 'surfactant_head_surfactant_head_rdf.txt',
                    'size_in_bytes': 115098,
                    'utc_timestamp': 1600729024.880694},
                '0b6c156fb7e71a1387904cc6d918cb9acf286538': {
                    'hash': 'bf9cb747ae826511958790034e62d773',
                    'relpath': 'counterion_counterion_rdf.txt',
                    'size_in_bytes': 115098,
                    'utc_timestamp': 1600729024.979725
                },
                ...
            }
        }

    Required params:
        - uri (str): dataset URI

    Optional params:
        - dest_dir (str): destination directory, default is the current working
            directory.
        - fizzle_empty_result (bool): fizzle if nothing found, default: True
        - manifest_file (str): default: None. If specififed, then manifest of
            dataset is written in .yaml or .json format, depending on the
            file name suffix.

        Fields 'uri', 'dest_dir', 'manifest_file',
        may also be a dict of format { 'key': 'some->nested->fw_spec->key' }
        for looking up value within fw_spec instead.
    """
    _fw_name = 'DirectManifestTask'
    required_params = [
        *DtoolLookupTask.required_params,
        "uri"]
    optional_params = [
        *DtoolLookupTask.optional_params,
        "dest_dir",
        "fizzle_empty_result",
        "manifest_file",
    ]

    def _run_task_internal(self, fw_spec):
        import dtoolcore
        logger = logging.getLogger(__name__)

        uri = self.get("uri", None)
        uri = from_fw_spec(uri, fw_spec)

        dest_dir = self.get("dest_dir", os.path.abspath(os.path.curdir))
        dest_dir = from_fw_spec(dest_dir, fw_spec)

        fizzle_empty_result = self.get("fizzle_empty_result", True)
        fizzle_empty_result = from_fw_spec(fizzle_empty_result, fw_spec)

        manifest_file = self.get("manifest_file", None)
        manifest_file = from_fw_spec(manifest_file, fw_spec)

        src_dataset = dtoolcore.DataSet.from_uri(uri)
        manifest = src_dataset.generate_manifest()
        if isinstance(manifest, str) and fizzle_empty_result:  # server error message
            raise ValueError(manifest)
        elif isinstance(manifest, str):
            logger.warning(manifest)
            manifest = None
        else:  # success
            logger.debug("Manifest of queried dataset:")
            _log_nested_dict(logger.debug, manifest)

            if manifest_file:
                manifest_file_path = os.path.join(dest_dir, manifest_file)
                logger.debug("Writing manifest of queried dataset to '{}'.".format(manifest_file_path))
                write_serialized(manifest, manifest_file_path)

        return manifest



#!/usr/bin/env python
#
# cmd_tasks.py
#
# Copyright (C) 2020 IMTEK Simulation
# Author: Johannes Hoermann, johannes.hoermann@imtek.uni-freiburg.de
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Tasks that modify the environment by lookups in the worker file."""

import logging
import io
import multiprocessing
import os
import pickle
import subprocess
import sys
import traceback
import threading

from contextlib import ExitStack, redirect_stdout, redirect_stderr
from queue import Queue, Empty
from six.moves import builtins

# fireworks-internal
from fireworks.core.firework import FiretaskBase, FWAction
from fireworks.user_objects.firetasks.script_task import ScriptTask, PyTask

# in order to have a somewhat centralized encoding configuration
from fireworks.utilities.fw_serializers import ENCODING_PARAMS

__author__ = 'Johannes Hoermann'
__copyright__ = 'Copyright 2018, IMTEK'
__version__ = '0.1.1'
__maintainer__ = 'Johannes Hoermann'
__email__ = 'johannes.hoermann@imtek.uni-freiburg.de'
__date__ = 'Mar 18, 2020'


# TODO: remove too verbose logging for pipes and streams

# TODO: own logging format on this level

# nice post on reading and writing streams:
# https://stackoverflow.com/questions/33395004/python-streamio-reading-and-writing-from-the-same-stream

# source: # https://stackoverflow.com/questions/4984428/python-subprocess-get-childrens-output-to-file-and-terminal
def tee(infile, *files):
    """Print `infile` to `files` in a separate thread."""
    def fanout(infile, *files):
        with infile:
            for line in iter(infile.readline, ''):
                for f in files:
                    f.write(line)

    t = threading.Thread(target=fanout, args=(infile,)+files)
    t.daemon = True
    t.start()
    return t


# NonBlockingStreamReader might be obsolete as long as all streams to be read
# are flushed and closed properly, eventually sending zero
# https://gist.github.com/EyalAr/7915597
class NonBlockingStreamReader:
    def __init__(self, stream):
        '''
        stream: the stream to read from.
                Usually a process' stdout or stderr.
        '''
        self.logger = logging.getLogger(__name__)
        self._s = stream
        self._q = Queue()

        def _populate_queue(stream, queue):
            '''
            Collect lines from 'stream' and put them in 'queue'.
            '''
            for line in iter(stream.readline, b''):
                if line is not None and len(line) > 0:
                    self.logger.debug("Queueing line '{}'.".format(line))
                queue.put(line)

        self._t = threading.Thread(target=_populate_queue,
            args=(self._s, self._q), name="NonBlockingStreamReader")
        self._t.daemon = True
        self._t.start()  # start collecting lines from the stream

    def readline(self, timeout=None):
        try:
            line = self._q.get(block=timeout is not None,
                timeout=timeout)
            if line is not None and len(line) > 0:
                self.logger.debug("Popped line '{}'.".format(line))
            return line
        except Empty:
            self.logger.debug("Nothing popped.")
            return None


# https://www.oreilly.com/library/view/python-cookbook/0596001673/ch06s03.html
class Tee(threading.Thread):
    """Tee of instream to an arbitrary number of outstreams."""
    def __init__(self, instream, *outstreams, **kwargs):
        """ constructor, setting initial variables """
        self.logger = logging.getLogger(__name__)
        self._stopevent = threading.Event()
        # self._sleepperiod = 1.0e-3  # s
        self.instream = instream
        self.outstreams = outstreams
        threading.Thread.__init__(self, **kwargs)

    def run(self):
        """ main control loop """
        # following line allows for this thread not to block when waiting
        # for data on a stream, but might also cause ending before the stream
        # is finished
        # instream = NonBlockingStreamReader(self.instream)
        instream = self.instream
        # finish reading in any case if there is still stuff coming through
        line = None
        # line is '' if stream ended
        while (not self._stopevent.isSet()) or (line != ''):
            if self._stopevent.isSet():
                # thread is about to stop, try to get remaining contents in stream
                # try to flush instream (if it's both readable AND writable)
                self.logger.debug(
                    "{} received stop signal, reading remaining stream data."
                        .format(self.name))
                try:
                    instream.flush()
                except Exception:
                    pass

            line = instream.readline()
            if line is not None:
                # if hasattr(self, 'name') and line is not None and len(line) > 0:
                #    self.logger.debug("{} captured '{}'".format(self.name, line))
                for f in self.outstreams:
                    f.write(line)
            # self._stopevent.wait(self._sleepperiod)

        for f in self.outstreams:
            f.flush()
        if hasattr(self, 'name'):
            self.logger.debug("{} received stop signal and finished.".format(self.name))

    def join(self, timeout=None):
        """ Stop the thread. """
        self._stopevent.set()
        if hasattr(self, 'name'):
            self.logger.debug("Ending thread '{}'".format(self.name))
        threading.Thread.join(self, timeout)


class TeeContext():
    """Tee context manager."""
    def __init__(self, instream, *outstreams, **kwargs):
        self.logger = logging.getLogger(__name__)
        self.instream = instream
        self.outstreams = outstreams
        self.tee = None
        self.kwargs = kwargs

    def __enter__(self):
        if 'name' in self.kwargs:
            self.logger.debug("Entering '{}' context.".format(self.kwargs['name']))
        self.tee = Tee(self.instream, *self.outstreams, **self.kwargs)
        self.tee.start()

    def __exit__(self, et, ev, tb):
        if 'name' in self.kwargs:
            self.logger.debug("Leaving '{}' context.".format(self.kwargs['name']))
        # self.tee.join(0.0)
        self.tee.join()


class PipeRWPair(io.BufferedRWPair):
    """Wraps io.BufferedRWPair around binary os.pipe."""
    def __init__(self, *args, name=None, **kwargs):
        self.logger = logging.getLogger(__name__)
        self.instance_name = name

        self._pipeout, self._pipein = os.pipe()
        if self.instance_name is not None:
            self.logger.debug(
                "{:s} opened pipe with infile {:d} and outfile {:d}."
                    .format(self.instance_name, self._pipein, self._pipeout))

        self.pipe_reader = io.FileIO(self._pipeout, mode='rb')
        self.pipe_writer = io.FileIO(self._pipein, mode='wb')
        super().__init__(self.pipe_reader, self.pipe_writer, *args, **kwargs)


class TextIOPipe(io.TextIOWrapper):
    """Wraps io.TextIOWrapper around PipeRWPair."""
    def __init__(self, *args, name=None, **kwargs):
        self.logger = logging.getLogger(__name__)
        self.instance_name = name

        self.pipe = PipeRWPair(name=name)
        super().__init__(self.pipe, *args, **kwargs)


class TextIOPipeContext(TextIOPipe):
    """Flushed underlying pipe writer when leaving context."""
    def __exit__(self, et, ev, tb):
        if self.instance_name is not None:
            self.logger.debug("Flushing {:s}.".format(self.instance_name))
        self.flush()


class LoggingContext():
    """Modifies logging within context."""
    def __init__(self, logger=None, handler=None, level=None, close=False):
        self.logger = logger
        self.level = level
        self.handler = handler
        self.close = close

    def __enter__(self):
        """Prepare log output streams."""
        if self.logger is None:
            # temporarily modify root logger
            self.logger = logging.getLogger('')
        if self.level is not None:
            self.old_level = self.logger.level
            self.logger.setLevel(self.level)
        if self.handler:
            self.logger.addHandler(self.handler)

    def __exit__(self, et, ev, tb):
        if self.level is not None:
            self.logger.setLevel(self.old_level)
        if self.handler:
            self.logger.removeHandler(self.handler)
        if self.handler and self.close:
            self.handler.close()
        # implicit return of None => don't swallow exceptions


# TODO: there seems to be an issue with unclean temporary environment
# at CmdTask only resolvable after rlaunch restart:
# NameError: name 'false' is not defined
class TemporaryOSEnviron:
    """Preserve original os.environ context manager."""

    def __enter__(self):
        """Store backup of current os.environ."""
        self._original_environ = os.environ.copy()

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Restore backed up os.environ."""
        os.environ = self._original_environ


# TODO: context for temporarily modified sys.path & sites, probably not perfect
class TemporarySysPath:
    """Preserve original os.environ context manager."""

    def __enter__(self):
        """Store backup of current sys.path."""
        self._original_sys_path = sys.path.copy()

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Restore backed up sys.path."""
        sys.path = self._original_sys_path


def get_nested_dict_value(d, key):
    """Uses '.'-splittable string as key to access nested dict."""
    if key in d:
        val = d[key]
    else:
        key = key.replace("->", ".")  # make sure no -> left
        split_key = key.split('.', 1)
        if len(split_key) == 2:
            key_prefix, key_suffix = split_key[0], split_key[1]
        else:  # not enough values to unpack
            raise KeyError("'{:s}' not in {}".format(key, d))

        val = get_nested_dict_value(d[key_prefix], key_suffix)

    return val


# https://stackoverflow.com/questions/19924104/python-multiprocessing-handling-child-errors-in-parent
class Process(multiprocessing.Process):
    """
    Class which returns child Exceptions to Parent.
    https://stackoverflow.com/a/33599967/4992248
    """

    def __init__(self, *args, **kwargs):
        multiprocessing.Process.__init__(self, *args, **kwargs)
        self._parent_conn, self._child_conn = multiprocessing.Pipe()
        self._exception = None

    def run(self):
        try:
            multiprocessing.Process.run(self)
            self._child_conn.send(None)
        except Exception as e:
            tb = traceback.format_exc()
            self._child_conn.send((e, tb))
            raise e  # You can still rise this exception if you need to

    @property
    def exception(self):
        if self._parent_conn.poll():
            self._exception = self._parent_conn.recv()
        return self._exception


class EnvTask(FiretaskBase):
    """Abstract base class for tasks that modify the environmen.

    All derivatives will have the option to
      * specify an environment 'env' to be looked up within the worker file,
      * redirect stdout, stderr and log streams to files 'stdout_file',
        'stderr_file', and 'stdlog_file'.
      * store stdout and stderr and log messages in database if 'stored_stdout',
        'store_stderr', 'store_stdlog' are set,
      * write simple (non-exhaustive) pseudo command history to 'py_hist_file'.

    Examples:

        For testing redirection of output and logging information to different
        locations, use a simple bash snippet `trial.sh`

            #!/bin/bash
            echo "This goes to stdout!"
            echo "This goes to stderr!" 1>&2

        and evoke via CmdTask with

        >>> from jlhfw.fireworks.user_objects.firetasks.cmd_tasks import CmdTask
        >>>
        >>> import logging
        >>>
        >>> logformat = (
        >>>             "[%(asctime)s-%(funcName)s()-%(filename)s:%(lineno)s]"
        >>>             " %(levelname)s: %(message)s")
        >>> logging.basicConfig(
        >>>     level=logging.DEBUG, format=logformat)
        >>>
        >>> task = CmdTask(
        >>>     cmd='bash',
        >>>     opt='trial.sh',
        >>>     stderr_file  = 'std.err',
        >>>     stdout_file  = 'std.out',
        >>>     stdlog_file  = 'std.log',
        >>>     py_hist_file = 'simple_hist.py',
        >>>     call_log_file= 'calls.log',
        >>>     vars_log_file= 'vars.log',
        >>>     store_stdout = True,
        >>>     store_stderr = True,
        >>>     store_stdlog = True,
        >>>     use_shell    = False,
        >>>     fizzle_bad_rc= True)
        >>>
        >>> fwaction = task.run_task({})
        >>> print(fwaction)

        and investigate the contents of 'std.err', 'std.out', 'std.log',
        'simple_hist.py', 'calls.log' and 'vars.log' as well as what you see
        as output on the terminal.
    """

    # required_params = []
    other_params = [
        # base class EnvTask params
        'env',
        'fork',  # fork new process before modifying environment if True
        'loglevel',
        'py_hist_file',
        'call_log_file', # obsolete
        'vars_log_file', # obsolete
        'stdout_file',
        'stderr_file',
        'stdlog_file',
        'store_stdout',
        'store_stderr',
        'store_stdlog',
        'propagate',  # propagate update_spec and mod_spec down all decendants
    ]

    def _py_hist_append(self, line):
        # self._py_hist_stream.write(line + '\n')
        if self._py_hist_logger is not None:
            self._py_hist_logger.info(line)

    def run_task(self, fw_spec):
        """Run a sub-process. Modify environment if desired."""
        if self.get('use_global_spec'):
            self._load_params(fw_spec)
        else:
            self._load_params(self)

        if self.fork:
            # spawn child process to assure my environment stays untouched
            q = multiprocessing.Queue()
            p = Process(target=self._run_task, args=(q, fw_spec,))
            p.start()

            # wait for child to queue fw_action object and
            # check whether child raises exception
            while q.empty():
                # if child raises exception, then it has terminated
                # before queueing any fw_action object
                if p.exception:
                    error, p_traceback = p.exception
                    raise ChildProcessError(p_traceback)
            # this loop will deadlock for any child that never raises
            # an exception and does not queue anything

            # child has finished without exception
            # child must always return fw_action
            # queue only used for one transfer of
            # return fw_action, should thus never deadlock.
            fw_action = q.get()
            # if we reach this line without the child
            # queueing anything, then process will deadlock.
            p.join()
        else:
            fw_action = self._run_task(None, fw_spec)

        return fw_action

    def _run_task(self, q, fw_spec):
        threads = []

        with ExitStack() as stack:
            # contexts to be entered before and left in reverse order after
            # task execution

            # logging to dedicated log stream if desired
            if self.store_stdlog:
                # stdlog_stream = io.TextIOWrapper(io.BytesIO(), **ENCODING_PARAMS)
                stdlog_stream = io.StringIO()
                logh = logging.StreamHandler(stdlog_stream)
                # optional 1st context: logging to in-memory buffer context
                stack.enter_context(
                    LoggingContext(handler=logh, level=self.loglevel, close=False))

            # logging to dedicated log file if desired
            if self.stdlog_file:
                logfh = logging.FileHandler(self.stdlog_file, mode='a', **ENCODING_PARAMS)
                # optional 2nd context: logging to file context
                stack.enter_context(
                    LoggingContext(handler=logfh, level=self.loglevel, close=True))

            # redirect stdout and stderr if desired (store them in db later)
            # create non-default streams to insert into db if desired
            out_streams = []
            if self.stdout_file:
                outf = open(self.stdout_file, 'a', **ENCODING_PARAMS)
                out_streams.append(outf)
            if self.store_stdout:
                stdout_stream = io.StringIO()  # io.TextIOWrapper(io.BytesIO(), **ENCODING_PARAMS)
                out_streams.append(stdout_stream)
            out_streams.append(sys.stdout)
            # stdout_pipe = TextIOPipeContext(name='StdOutPipe', **ENCODING_PARAMS)
            stdout_pipeout, stdout_pipein = os.pipe()

            # 3rd context: stdout tee
            stdout_reader = stack.enter_context(
                io.TextIOWrapper(io.FileIO(stdout_pipeout, mode='rb'), **ENCODING_PARAMS))

            stack.enter_context(
                TeeContext(stdout_reader, *out_streams, name='StdOutTee'))

            # 4th context: stdout pipe
            # stack.enter_context(stdout_pipe)

            err_streams = []
            if self.stderr_file:
                errf = open(self.stderr_file, 'a', **ENCODING_PARAMS)
                err_streams.append(errf)
            if self.store_stderr:
                stderr_stream = io.StringIO()
                err_streams.append(stderr_stream)
            err_streams.append(sys.stderr)
            # stderr_pipe = TextIOPipeContext(name='StdErrPipe', **ENCODING_PARAMS)
            stderr_pipeout, stderr_pipein = os.pipe()

            # 5th context: stderr tee
            stderr_reader = stack.enter_context(
                io.TextIOWrapper(io.FileIO(stderr_pipeout, mode='rb'), **ENCODING_PARAMS))

            stack.enter_context(
                TeeContext(stderr_reader, *err_streams, name='StdErrTee'))

            # 6th context stderr pipe
            # stack.enter_context(stderr_pipe)

            # always "tunnel" stdout and stderr through our buffers for
            # arbitrary tee'ing and capturing
            # 7th context: redirect sys.stdout into pipe
            # self.pipe_writer = io.FileIO(self._pipein, mode='wb')
            stdout_writer = stack.enter_context(
                io.TextIOWrapper(io.FileIO(stdout_pipein, mode='wb'), **ENCODING_PARAMS))

            stack.enter_context(redirect_stdout(stdout_writer))
            # stack.enter_context(redirect_stdout(stdout_pipe))
            # 8th context: redirect sys.stderr into pipe
            # stack.enter_context(redirect_stderr(stderr_pipe))
            stderr_writer = stack.enter_context(
                io.TextIOWrapper(io.FileIO(stderr_pipein, mode='wb'), **ENCODING_PARAMS))

            stack.enter_context(redirect_stderr(stderr_writer))

            # write pseudo command history to file if desired
            if self.py_hist_file:
                py_hist_handler = logging.FileHandler(self.py_hist_file, mode='a', **ENCODING_PARAMS)
                py_hist_formatter = logging.Formatter('%(message)s')
                py_hist_handler.setFormatter(py_hist_formatter)
            else:
                py_hist_handler = logging.NullHandler()
            self._py_hist_logger = logging.getLogger(__name__ + '.py_hist_logger')
            self._py_hist_logger.propagate = False
            self._py_hist_logger.setLevel(logging.DEBUG)
            for h in self._py_hist_logger.handlers:
                self._py_hist_logger.removeHandler(h)
            self._py_hist_logger.addHandler(py_hist_handler)

            # log, stdout, stderr streams are logged to
            # files and database, depending on flags.

            # 11th context: allow temporary os.environ modifications within block
            stack.enter_context(TemporaryOSEnviron())
            # 12th context: allow temporary sys.path modifications within block
            stack.enter_context(TemporarySysPath())

            # for a variable amount of contexts:
            # for mgr in context_managers:
            #    stack.enter_context(mgr)
            # TODO: replace own logger attribute with getLogger in methods
            self.logger = logging.getLogger(__name__)
            ret = self._run_task_internal(fw_spec)

        for thread in threads:
            thread.join()  # wait for stream tee threads

        if isinstance(ret, FWAction):
            fw_action = ret
        else:
            fw_action = FWAction()

        output = {}
        if self.store_stdout:
            stdout_stream.flush()
            # stdout_stream.seek(0)
            # output['stdout'] = stdout_stream.read()
            output['stdout'] = stdout_stream.getvalue()

        if self.store_stderr:
            stderr_stream.flush()
            # stderr_stream.seek(0)
            # output['stderr'] = stderr_stream.read()
            output['stderr'] = stderr_stream.getvalue()

        if self.store_stdlog:
            stdlog_stream.flush()
            # stdlog_stream.seek(0)
            # output['stdlog'] = stdlog_stream.read()
            output['stdlog'] = stdlog_stream.getvalue()

        fw_action.stored_data.update(output)

        # 'propagate' only development feature for now
        if hasattr(fw_action, 'propagate') and self.propagate is not None:
            fw_action.propagate = self.propagate

        # return fw_action
        if self.fork:
            q.put(fw_action)
        else:
            return fw_action

    # TODO: get rid of those _load_params functions (relic from ScriptTask)
    def _load_params(self, d):
        self.env = self.get('env')
        self.fork = self.get('fork', False)
        self.loglevel = self.get('loglevel', logging.DEBUG)
        self.py_hist_file = self.get('py_hist_file', 'simple_hist.py')

        self.stdout_file = d.get('stdout_file', 'std.out')
        self.stderr_file = d.get('stderr_file', 'std.err')
        self.stdlog_file = d.get('stdlog_file', 'std.log')
        self.store_stdout = d.get('store_stdout')
        self.store_stderr = d.get('store_stderr')
        self.store_stdlog = d.get('store_stdlog')

        self.propagate = d.get('propagate')


class CmdTask(EnvTask, ScriptTask):
    """Enhanced script task, runs (possibly environment dependent)  command.

    Required params:
        - cmd (str): command to look up within '_fw_env (see below for details)

    Optional params:
        - opt ([obj]): list of strings (or int, float, ...) to pass as
            options / arguments tp 'cmd'. List may contain dicts of
            pattern {'key': 'some->nested->fw_spec->key'} or
            {'eval': 'python-expression'} in order to dynamically modify opt.
        - env (str): allows to specify an environment possibly defined in the
            worker file. If so, additional environment-related iintialization
            and expansion of command aliases are carried out (see below).
        - defuse_bad_rc - (default:False) - if set True, a non-zero returncode
            from the Script (indicating error) will cause FireWorks to defuse
            the child FireWorks rather than continuing.
        - fizzle_bad_rc - (default:False) - if set True, a non-zero returncode
            from the Script (indicating error) will cause the Firework to raise
            an error and FIZZLE.
        - use_shell - (default:True) - whether to execute the command through
            the current shell (e.g., BASH or CSH). If true, you will be able
            to use environment variables and shell operators; but, this method
            can be less secure.
        - shell_exe - (default:None) - shell executable, e.g. /bin/bash.
            Generally, you do not need to set this unless you want to run
            through a non-default shell.
        - stdin_file - (default:None) - feed this filepath as standard input
            to the script
        - stdin_key - (default:None) - feed this string or list of strings
            as standard input to the script
        - store_stdout (default:False) - store the entire standard output in
            the Firework Launch object's stored_data
        - stdout_file - (default:None) - store the entire standard output in
            this filepath. If None, the standard out will be streamed to
            sys.stdout
        - store_stderr - (default:False) - store the entire standard error in
            the Firework Launch object's stored_data
        - stderr_file - (default:None) - store the entire standard error in
            this filepath. If None, the standard error will be streamed to
            sys.stderr.
        - propagate (bool, default:None): if True, then set the
            FWAction 'propagate' flag and propagate all 'update_spec' and
            'mod_spec' not only to direct children, but to all descendants
            down to the wokflow's leaves.

    Makes use of '_fw_env' worker specific definitions and allow for certain
    abstraction when running commands, i.e.

        - _fw_name: CmdTask
          cmd: lmp
          opt:
          - -in lmp_production.input
          - -v surfactant_name SDS
          - -v has_indenter 1
          - -v constant_indenter_velocity -0.0001
          - -v productionSteps 375000
          [...]
          - -v store_forces 1
          stderr_file:   std.err
          stdout_file:   std.out
          fizzle_bad_rc: true
          use_shell:     true

    will actually look for a definition 'lmp' within a worker's '_fw_env' and
    if available execute in place of the  specififed command, simply appending
    the list of options given in 'opt' separated by sapces. This allows
    machine-specific cod to be placed within the worker files, i.e. for a
    machine called NEMO within nemo_queue_worker.yaml

        name: nemo_queue_worker
        category: [ 'nemo_queue' ]
        query: '{}'
        env:
          lmp: module purge;
               module use /path/to/modulefiles;
               module use ${HOME}/modulefiles;
               module load lammps/16Mar18-gnu-7.3-openmpi-3.1-colvars-09Feb19;
               mpirun ${MPIRUN_OPTIONS} lmp

    and for a machine called JUWELS within

        name:     juwels_queue_worker
        category: [ 'juwels_queue' ]
        query:    '{}'
        env:
          lmp:  module purge;
                jutil env activate -p chfr13;
                module use ${PROJECT}/hoermann/modules/modulefiles;
                module load Intel/2019.0.117-GCC-7.3.0 IntelMPI/2018.4.274;
                module load jlh/lammps/16Mar18-intel-2019a
                srun lmp

    A third machine's worker file might look like this:

        name: bwcloud_std_fworker
        category: [ 'bwcloud_std', 'bwcloud_noqueue' ]
        query: '{}'
        env:
          lmp:  module purge;
                module load LAMMPS;
                mpirun -n 4 lmp
          vmd:  module purge;
                module load VMD;
                vmd
          pizza.py: module load MDTools/jlh-25Jan19-python-2.7;
                    pizza.py

    This allows for machine-independent workflow design.

    More sophisticated environment modifications are possible when
    explicitly specifying the 'env' option. Worker files can contain

        env:
            python:
                init:
                - 'import sys, os'
                - 'sys.path.insert(0, os.path.join(os.environ["MODULESHOME"], "init"))'
                - 'from env_modules_python import module'
                - 'module("use","/path/to/modulefiles")'
                cmd:
                    gmx:
                        init: 'module("load","GROMACS/2019.3")'
                        substitute: gmx_mpi
                        prefix: ['mpirun', { 'eval': 'os.environ["MPIRUN_OPTIONS"]' } ]

    corresponds to a bash snippet

        module use /path/to/modulefiles
        module load GROMACS/2019.3
        mpirun ${MPIRUN_OPTIONS} gmx_mpi ...

    extended by arguments within the task's 'opt' parameter.

    """
    _fw_name = 'CmdTask'
    required_params = ['cmd']
    optional_params = [
        # base class EnvTask params
        *EnvTask.other_params,
        # CmdTask params
        'opt',
        'stdin_file',
        'stdin_key',
        'use_shell',
        'shell_exe',
        'defuse_bad_rc',
        'fizzle_bad_rc'
    ]

    @property
    def args(self):
        """List of arguments (including command) as list of str only."""
        return [a if isinstance(a, str) else str(a) for a in self._args]

    def _run_task_internal(self, fw_spec):
        """Run task."""
        stdout = subprocess.PIPE # if self.store_stdout or self.stdout_file else None
        stderr = subprocess.PIPE # if self.store_stderr or self.stderr_file else None

        # get the standard in and run task internally
        if self.stdin_file:
            stdin = open(self.stdin_file, 'r', **ENCODING_PARAMS)
        elif self.stdin_key:
            stdin_str = fw_spec[self.stdin_key]
            try:
                stdin_list = [stdin_str] if isinstance(stdin_str, str) \
                    else [str(line) for line in stdin_str]
            except:
                raise ValueError(("stdin_key '{}' must point to either string"
                                  " or list of strings, not '{}'.")
                                  .format(self.stdin_key, stdin_str))

            stdin = subprocess.PIPE
        else:
            stdin = None

        ### START WRITING PY_HIST
        self._py_hist_append('import os')
        self._py_hist_append('import subprocess')

        self._args = []
        # in case of a specified worker environment

        # in case of a specified worker environment
        if self.env and "_fw_env" in fw_spec \
                and self.env in fw_spec["_fw_env"]:
            self.logger.info("Found {:s}-specific block '{}' within worker file."
                .format(self.env, fw_spec["_fw_env"]))

            # """Parse global init block."""
            # _fw_env : env : init may provide a list of python commans
            # to run, i.e. for module env initialization
            if "init" in fw_spec["_fw_env"][self.env]:
                init = fw_spec["_fw_env"][self.env]["init"]
                if isinstance(init, str):
                    init = [init]
                assert isinstance(init, list)
                for cmd in init:
                    self.logger.info("Execute '{:s}'.".format(cmd))
                    self._py_hist_append(cmd)
                    exec(cmd)

            # """Parse global envronment block."""
            # per default, process inherits current environment
            # self.modenv = os.environ.copy()
            # modify environment before call if desired
            if "env" in fw_spec["_fw_env"][self.env]:
                env_dict = fw_spec["_fw_env"][self.env]["env"]
                if not isinstance(env_dict, dict):
                    raise ValueError(
                        "type({}) = {} of 'env' not accepted, must be dict!"
                            .format(env_dict, type(env_dict)))

                # so far, only simple overrides, no pre- or appending
                for i, (key, value) in enumerate(env_dict.items()):
                    self.logger.info("Set env var '{:s}' = '{:s}'.".format(
                        key, value))
                    self._py_hist_append('os.environ["{:s}"] = "{:s}"'.format(
                        str(key), str(value)))
                    os.environ[str(key)] = str(value)

        if self.env and "_fw_env" in fw_spec \
                and self.env in fw_spec["_fw_env"]:
            # check whether there is any machine-specific "expansion" for
            # the command head
            """Parse command-specific environment block."""
            if "cmd" in fw_spec["_fw_env"][self.env] \
                    and self.cmd in fw_spec["_fw_env"][self.env]["cmd"]:
                self.logger.info("Found {:s}-specific block '{}' within worker file."
                    .format(self.cmd, fw_spec["_fw_env"][self.env]["cmd"][self.cmd]))
                # same as above, evaluate command-specific initialization code
                """Parse per-command init block."""
                cmd_block = fw_spec["_fw_env"][self.env]["cmd"][self.cmd]
                if "init" in cmd_block:
                    init = cmd_block["init"]
                    if isinstance(init, str):
                        init = [init]
                    assert isinstance(init, list), "'init' must be str or list"
                    for cmd in init:
                        self.logger.info("Execute '{:s}'.".format(cmd))
                        self._py_hist_append(cmd)
                        exec(cmd)

                """Parse per-command substitute block."""
                cmd_block = fw_spec["_fw_env"][self.env]["cmd"][self.cmd]
                if "substitute" in cmd_block:
                    substitute = cmd_block["substitute"]
                    assert isinstance(substitute, str), "substitute must be str"
                    self.logger.info("Substitute '{:s}' with '{:s}'.".format(
                        self.cmd, substitute))
                    self._args.append(substitute)
                else:  # otherwise just use command as specified
                    self.logger.info("No substitute for command '{:s}'.".format(
                        self.cmd))
                    self._args.append(self.cmd)

                """Parse per-command prefix block."""
                cmd_block = fw_spec["_fw_env"][self.env]["cmd"][self.cmd]
                # prepend machine-specific prefixes to command
                if "prefix" in cmd_block:
                    prefix_list = cmd_block["prefix"]
                    if not isinstance(prefix_list, list):
                        prefix_list = [prefix_list]

                    processed_prefix_list = []
                    for i, prefix in enumerate(prefix_list):
                        processed_prefix = []
                        # a prefix dict allow for something like this:
                        #    cmd:
                        #      lmp:
                        #        init:   'module("load","lammps")'
                        #        prefix: [ 'mpirun', { 'eval': 'os.environ["MPIRUN_OPTIONS"]' } ]
                        if isinstance(prefix, dict):
                            # special treatment desired for this prefix
                            if "eval" in prefix:
                                self.logger.info("Evaluate prefix '{:s}'.".format(
                                    ' '.join(prefix["eval"])))
                                # evaluate prefix in current context
                                processed_prefix = eval(prefix["eval"])
                                try:
                                    processed_prefix = processed_prefix.decode("utf-8")
                                except AttributeError:
                                    pass
                                if isinstance(processed_prefix, str):
                                    processed_prefix = processed_prefix.split()
                                else:
                                    raise ValueError(
                                        "Output {} of prefix #{} evaluation not accepted!".format(
                                            processed_prefix, i))
                            else:
                                raise ValueError(
                                    "Formatting {} of prefix #{} not accepted!".format(
                                        prefix, i))
                        elif isinstance(prefix, str):
                            # prefix is string, not much to do, split & prepend
                            processed_prefix = prefix.split()
                        else:
                            raise ValueError(
                                "type({}) = {} of prefix #{} not accepted!".format(
                                    prefix, type(prefix), i))

                        if not isinstance(processed_prefix, list):
                            processed_prefix = [processed_prefix]

                        self.logger.info("Prepend prefix '{:s}'.".format(
                            ' '.join(processed_prefix)))
                        processed_prefix_list.extend(processed_prefix)

                    self._args = processed_prefix_list + self._args  # concat two lists

                """Parse command-specific environment block."""
                # per default, process inherits current environment

                cmd_block = fw_spec["_fw_env"][self.env]["cmd"][self.cmd]
                # modify environment before call if desired
                if "env" in cmd_block:
                    env_dict = cmd_block["env"]
                    if not isinstance(env_dict, dict):
                        raise ValueError(
                            "type({}) = {} of 'env' not accepted, must be dict!"
                            .format(env_dict, type(env_dict)))

                    # so far, only simple overrides, no pre- or appending
                    for i, (key, value) in enumerate(env_dict.items()):
                        self.logger.info("Set env var '{:s}' = '{:s}'.".format(
                            key, value))
                        self._py_hist_append('os.environ["{:s}"] = "{:s}"'.format(
                            str(key), str(value)))
                        os.environ[str(key)] = str(value)
            else:  # no command-specific expansion in environment, use as is
                self._args.append(self.cmd)
        elif "_fw_env" in fw_spec and self.cmd in fw_spec["_fw_env"]:
            # check whether there is any desired command and whether there
            # exists a machine-specific "alias"
            self.logger.info("Found root-level {:s}-specific block '{}' within worker file."
                .format(self.cmd, fw_spec["_fw_env"][self.cmd]))
            self._args = [fw_spec["_fw_env"][self.cmd]]
        else:
            self._args = [self.cmd]

        if self.opt:  # extend by command line options if available
            # evaluate raplacement rules, i.e. fw_spec key or expression evaluation
            for arg in self.opt:
                if isinstance(arg, dict):
                    # only one rule per opt item allowed
                    # insert value from (nested) key - value pair in fw_spec
                    if 'key' in arg:
                        arg = get_nested_dict_value(fw_spec, arg['key'])
                    # evaluate expression and insert result
                    elif 'eval' in arg:
                        arg = eval(arg['eval'])
                # user should only use strings as args
                # to be sure, always append string representation
                self._args.append(str(arg))

        self.logger.info("Built args '{}'.".format(self._args))
        self.logger.info("Built command '{:s}'.".format(
            ' '.join(self.args)))
        self.logger.debug("Process environment '{}'.".format(os.environ))

        kwargs = {}
        if self.shell_exe:
            kwargs.update({'executable': self.shell_exe})
        if self.use_shell:
            kwargs.update({'shell': self.use_shell})

        # Write .py script to reconstruct environment, very rudimentary.
        # The idea of this "history" file (just text written to a file,
        # no actual history) is to be able to reproduce runs quickly
        # outside of the FireWorks frameworks for debugging purposes.
        encoding_params_str_list = [
            '{}="{}"'.format(k, v) if isinstance(v, str)
            else '{}={}'.format(k, v) for k, v in ENCODING_PARAMS.items()]
        encoding_params_str = ', '.join(encoding_params_str_list)

        kwargs_str_list = [
            '{}="{}"'.format(k, v) if isinstance(v, str)
            else '{}={}'.format(k, v) for k, v in kwargs.items()]
        kwargs_str = ', '.join(kwargs_str_list)

        self._py_hist_append('# os.environ = {}'.format(dict(os.environ)))

        stdoutstr = 'subprocess.PIPE' # if self.store_stdout or self.stdout_file else 'None'
        stderrstr = 'subprocess.PIPE' # if self.store_stderr or self.stderr_file else 'None'
        if self.stdin_key:
            stdinstr = 'subprocess.PIPE'
        elif self.stdin_file:
            stdinstr = 'open("{:s}", "r", {:s})'.format(
                self.stdin_file, encoding_params_str)
        else:
            stdinstr = 'None'

        self._py_hist_append('')
        self._py_hist_append('p = subprocess.Popen(')
        self._py_hist_append('    {},'.format(self.args))
        self._py_hist_append('    stdin={},'.format(stdinstr))
        self._py_hist_append('    stdout={},'.format(stdoutstr))
        self._py_hist_append('    stderr={},'.format(stderrstr))
        self._py_hist_append('    env=os.environ,')
        if len(encoding_params_str) > 0:
            self._py_hist_append('    {},'.format(encoding_params_str))
        if len(kwargs) > 0:
            self._py_hist_append('    {},'.format(kwargs_str))
        self._py_hist_append(')')
        self._py_hist_append('')

        if self.stdin_key:
            self._py_hist_append('p.stdin.writelines({})'.format(stdin_list))
            self._py_hist_append('p.stdin.close()')

        self._py_hist_append('ret = p.wait()')
        ### DONE WRITING PY_HIST

        self.logger.info("Evoking subprocess.Popen with...")
        self.logger.info("    args            '{}'.".format(self.args))
        self.logger.info("    ENCODING_PARAMS '{}'.".format(ENCODING_PARAMS))
        self.logger.info("    kwargs          '{}'.".format(kwargs))

        p = subprocess.Popen(
            self.args, stdin=stdin,
            stdout=stdout, stderr=stderr, env=os.environ,
            **ENCODING_PARAMS, **kwargs)

        self.logger.info("Evoked process with...")
        self.logger.info("    args         '{}'.".format(p.args))
        self.logger.info("    PID          '{}'.".format(p.pid))
        self.logger.info("    type(stdin)  '{}'.".format(type(p.stdin)))
        self.logger.info("    type(stdout) '{}'.".format(type(p.stdout)))
        self.logger.info("    type(stderr) '{}'.".format(type(p.stderr)))

        threads = []

        out_streams = []
        out_streams.append(sys.stdout)  # per default to sys.stdout
        threads.append(tee(p.stdout, *out_streams))

        err_streams = []
        err_streams.append(sys.stderr)  # per default to sys.stderr
        threads.append(tee(p.stderr, *err_streams))

        # send stdin if desired and wait for subprocess to complete
        if self.stdin_key:
            p.stdin.writelines(stdin_list)

        try:
            p.stdin.close()
        except:
            pass

        for thread in threads:
            thread.join()  # wait for stream tee threads
        # tee threads close stderr and stdout

        returncode = p.wait()
        self.logger.info(
            "Process returned '{}'.".format(returncode))

        output = {}

        output['returncode'] = returncode

        if self.defuse_bad_rc and returncode != 0:
            fwaction = FWAction(stored_data=output, defuse_children=True)
        elif self.fizzle_bad_rc and returncode != 0:
            raise RuntimeError(
                'CmdTask fizzled! Return code: {}, output: {}'
                .format(returncode, output))
        else:  # returned as expected with 0 return code
            fwaction = FWAction(stored_data=output)

        return fwaction

    def _load_params(self, d):
        """Load parameters from task specs into attributes."""
        EnvTask._load_params(self, d)

        if d.get('stdin_file') and d.get('stdin_key'):
            raise ValueError('CmdTask cannot process both a key and file as the standard in!')

        self.cmd = d.get('cmd')

        # command line options
        opt = d.get('opt', None)
        # opt may be mixed list of [str,dict]
        if (opt is not None) and (not isinstance(opt, list)):
            opt = [opt]  # no check on str or dict elements!

        self.opt = opt

        self.stdin_file = d.get('stdin_file')
        self.stdin_key = d.get('stdin_key')

        self.shell_exe = d.get('shell_exe')
        self.use_shell = d.get('use_shell')
        self.defuse_bad_rc = d.get('defuse_bad_rc')
        self.fizzle_bad_rc = d.get('fizzle_bad_rc', not self.defuse_bad_rc)

        if self.defuse_bad_rc and self.fizzle_bad_rc:
            raise ValueError(
                'CmdTask cannot both FIZZLE and DEFUSE a bad returncode!')


class PyEnvTask(EnvTask, PyTask):
    """Same as PyTask, but allows to modify environment and inject snippets.

    First, performs environment lookup in worker file by `env` (see CmdTask).
    Second, runs python code lines in `init` before calling `func`.
    Offers same logging functionalyte as CmdTask. The optional inputs
    and outputs lists may contain spec keys to add to args list and to make
    the function output available in the curent and in children fireworks.
    Required parameters:
        - func (str): Fully qualified python method. E.g., json.dump, or shutil
            .copy, or some other function that is not part of the standard
            library!
    Optional parameters:
        - init: (str or [str]): Python code to execute before calling `func`,
            i.e. function definitions or imports. May access `fw_spec`.
        - args (list): List of args. Default is empty.
        - kwargs (dict): Dictionary of keyword args. Default is empty.
        - auto_kwargs (bool): If True, all other params not starting with '_'
            are supplied as keyword args.
        - kwargs_inputs (dict): dict of keyword argument names and
            possibly nested fw_spec keys for looking up inputs dynamically.
            Relates to `kwargs` like `inputs` relates to `args`, but the
            generated keyword arguments dict will override static key - value
            pairs in kwargs. fwspec keys can be '.' or '->'-delimited key
            to nested fields.
        - stored_data_varname (str): Whether to store the output in FWAction.
            If this is a string that does not evaluate to False, the output of
            the function will be stored as
            FWAction(stored_data={stored_data_varname: output}). The name is
            deliberately long to avoid potential name conflicts.
        - inputs ([str]): a list of keys in spec which will be used as inputs;
            the generated arguments list will be appended to args.
            Can be key '.' or '->'-delimited key to nested fields.
        - outputs ([str]): a list of spec keys that will be used to pass
            the function's outputs to child fireworks.
        - env (str): allows to specify an environment possibly defined in the
            worker file. If so, additional environment-related intialization
            carried out (see CmdTask).
        - store_stdout (default:False) - store the entire standard output in
            the Firework Launch object's stored_data.
        - stdout_file - (default:None) - store the entire standard output in
            this filepath. If None, the standard out will be streamed to
            sys.stdout.
        - store_stderr - (default:False) - store the entire standard error in
            the Firework Launch object's stored_data.
        - stderr_file - (default:None) - store the entire standard error in
            this filepath. If None, the standard error will be streamed to
            sys.stderr.
        - chunk_number (int): a serial number of the Firetask within a
            group of Firetasks generated by a ForeachTask.
        - propagate (bool, default:None): if True, then set the
            FWAction 'propagate' flag and propagate outputs not only to
            direct children, but all descendants down to the wokflow's leaves.
            If set, then any 'propagate' flag possibly set by an evoked
            function that returns FWAction.

    Examples:
        In order to quickly use a python snippet that we cannot expect to
        be available as a fully qualified function at the worker, we could do

        >>> def get_element_name(n):
        >>>     import ase.data
        >>>     return ase.data.atomic_names[n]
        >>>
        >>> import dill
        >>> func_str = dill.dumps(get_element_name)
        >>> init_lst = [
        >>>     'import builtins, dill',
        >>>     'builtins.injected_func = dill.loads({})'.format(func_str) ]
        >>>
        >>> ft = PyEnvTask(init=init_lst, func='injected_func',
        >>>                args=[18], outputs=['element_name'])
        >>>

        as long as at least 'dill' is available at the worker. The result is

        >>> fw_action = ft.run_task({})
        >>> print(fw_action.as_dict())
        {'stored_data': {}, 'exit': False,
            'update_spec': {'element_name': 'Argon'}, 'mod_spec': [],
            'additions': [], 'detours': [], 'defuse_children': False,
            'defuse_workflow': False}

        The injection into 'builtin' is necessary as evaluating 'init' and
        evoking 'func' won't happen within the same local scope.

        In order to make some Python package available that requires
        the loading of an environmnet module, the environment-specific
        snippet can be stored within a worker file, i.e.

          name:     juwels_noqueue_worker
          category: [ 'juwels_noqueue' ]
          query:    '{}'
          env:
            modified_py_env:
              init:
              - 'import site, sys, os, importlib, builtins'
              - 'sys.path.insert(0, os.path.join(os.environ["MODULESHOME"], "init"))'
              - 'builtins.module = importlib.import_module("env_modules_python").module'
              - 'module("purge")'
              - 'module("use",os.path.join(os.environ["PROJECT"],"common","juwels","easybuild","otherstages"))'
              - 'module("load","Stages/2019a","Intel/2019.3.199-GCC-8.3.0","IntelMPI/2019.3.199")'
              - 'module("load","ASE/3.17.0-Python-3.6.8")'
              - 'module("load","imteksimcs/devel-local-Python-3.6.8")'
              - 'module("load","imteksimpyenv/devel-Python-3.6.8")'
              - 'for d in list(set(os.environ["PYTHONPATH"].split(":")) - set(sys.path)): site.addsitedir(d)'

        then looked up and executed by PyEnvTask before the actual call to 'func'
        if the task's parameter 'env' points to 'mod_py_env'.
    """
    # I would prefer a plain-text injection, but 'dill' does the job.

    _fw_name = 'PyEnvTask'
    required_params = ['func']
    other_params = [
        # base class EnvTask params
        *EnvTask.other_params,
        *PyTask.other_params,
        'kwargs_inputs',
        # PyEnvTask params
        'init'
    ]
    # TODO: remove 'other_params' to make this comment valid again:
    # note that we are not using "optional_params" because we do not want to do
    # strict parameter checking in FireTaskBase due to "auto_kwargs" option

    def _get_func(self, fw_spec):
        """Get function from string."""
        # self.logger.info("type('func') == str.")
        toks = self['func'].rsplit('.', 1)

        if len(toks) == 2:
            modname, funcname = toks
            self._py_hist_append('from {} import {} as func'
                .format(modname, funcname))
            self.logger.info(
                "'func' is a fully qualified name, 'from {} import {}'"
                    .format(modname, funcname))
            mod = __import__(modname, globals(), locals(), [str(funcname)], 0)
            func = getattr(mod, funcname)
        else:
            # Handle built in functions.
            self._py_hist_append('func = {}'.format(toks[0]))
            self.logger.info(
                "'func' is an unqualified name '{}', call directly."
                    .format(toks[0]))
            func = getattr(builtins, toks[0])
        return func

    def _run_task_internal(self, fw_spec):
        # run snippet
        # in case of a specified worker environment
        if self.env and "_fw_env" in fw_spec \
                and self.env in fw_spec["_fw_env"]:
            self.logger.info("Found {:s}-specific block '{}' within worker file."
                .format(self.env, fw_spec["_fw_env"]))

            # """Parse global init block."""
            # _fw_env : env : init may provide a list of python commans
            # to run, i.e. for module env initialization
            if "init" in fw_spec["_fw_env"][self.env]:
                init = fw_spec["_fw_env"][self.env]["init"]
                if isinstance(init, str):
                    init = [init]
                assert isinstance(init, list)
                for cmd in init:
                    self.logger.info("Execute '{:s}'.".format(cmd))
                    self._py_hist_append(cmd)
                    exec(cmd)

            # """Parse global envronment block."""
            # per default, process inherits current environment
            # self.modenv = os.environ.copy()
            # modify environment before call if desired
            if "env" in fw_spec["_fw_env"][self.env]:
                env_dict = fw_spec["_fw_env"][self.env]["env"]
                if not isinstance(env_dict, dict):
                    raise ValueError(
                        "type({}) = {} of 'env' not accepted, must be dict!"
                            .format(env_dict, type(env_dict)))

                # so far, only simple overrides, no pre- or appending
                for i, (key, value) in enumerate(env_dict.items()):
                    self.logger.info("Set env var '{:s}' = '{:s}'.".format(
                        key, value))
                    self._py_hist_append('os.environ["{:s}"] = "{:s}"'.format(
                        str(key), str(value)))
                    os.environ[str(key)] = str(value)

        # """Parse task-internal preceding code block."""
        if self.init:
            init = self.init
            if isinstance(init, str):
                init = [init]
            assert isinstance(init, list)
            for cmd in init:
                self.logger.info("Execute '{:s}'.".format(cmd))
                self._py_hist_append(cmd)
                exec(cmd)

        func = self._get_func(fw_spec)

        assert callable(func), ("Evaluated 'func' is {}, must be 'callable'"
            .format(type(func)))

        args = list(self.get('args', []))  # defensive copy

        self.logger.info("'args = {}'".format(args))

        inputs = self.get('inputs', [])
        self.logger.info("'inputs = {}'".format(inputs))

        assert isinstance(inputs, list), "'inputs' must be list."
        for item in inputs:
            args.append(get_nested_dict_value(fw_spec, item))

        if self.get('auto_kwargs'):
            kwargs = {k: v for k, v in self.items()
                      if not (k.startswith('_')
                              or k in self.required_params
                              or k in self.other_params)}
        else:
            kwargs = self.get('kwargs', {})

        self.logger.info("'kwargs = {}'".format(kwargs))

        kwargs_inputs = self.get('kwargs_inputs', {})
        self.logger.info("'kwargs_inputs = {}'".format(kwargs_inputs))
        assert isinstance(kwargs_inputs, dict), "'kwargs_inputs' must be dict."
        for kwarg_name, kwarg_key in kwargs_inputs.items():
            kwargs[kwarg_name] = get_nested_dict_value(fw_spec, kwarg_key)

        if len(args) > 0:
            self._py_hist_append('args = {}'.format(args))
        else:
            self._py_hist_append('args = []')

        if len(kwargs) > 0:
            self._py_hist_append('kwargs = {}'.format(kwargs))
        else:
            self._py_hist_append('kwargs = {}')

        self._py_hist_append('output = func(*args, **kwargs)')

        output = func(*args, **kwargs)

        if isinstance(output, FWAction):
            self.logger.info("'type(output) == FWAction', return directly.")
            return output

        self.logger.info("'type(output) == {}', build FWAction."
            .format(type(output)))
        actions = {}
        outputs = self.get('outputs', [])
        assert isinstance(outputs, list)
        if len(outputs) == 1:
            if self.get('chunk_number') is None:
                # actions['update_spec'] = {outputs[0]: output}
                actions['mod_spec'] = [{'_set': {outputs[0]: output}}]
            else:
                if isinstance(output, (list, tuple, set)):
                    mod_spec = [{'_push': {outputs[0]: i}} for i in output]
                else:
                    mod_spec = [{'_push': {outputs[0]: output}}]
                actions['mod_spec'] = mod_spec
        elif len(outputs) > 1:
            # assert isinstance(output, (list, tuple, set))  # fails in case of numpy array
            assert len(output) == len(outputs) # implies above assertion and should incorporate more iterable types
            # actions['update_spec'] = dict(zip(outputs, output))
            actions['mod_spec'] = [{'_set': dict(zip(outputs, output))}]

        if self.get('stored_data_varname'):
            actions['stored_data'] = {self['stored_data_varname']: output}
        if len(actions) > 0:
            self.logger.info("Built actions '{}.'".format(actions))
            return FWAction(**actions)

    def _load_params(self, d):
        """Load parameters from task specs into attributes."""
        EnvTask._load_params(self, d)
        # PyTask does not need to load params (?)
        self.init = self.get('init', None)


class PickledPyEnvTask(PyEnvTask):
    """Same as PyEnvTask, but expects pickled function instead of function name.

    As bytes might actually be stored as their bytes literals string
    representation within FireWork's, the task will try to eval 'func'
    before unpickling it if its type is str and not bytes.

    Examples:
        The shorter equivalent to above's PyEnvTask example looks likes this:

        >>> func_str = dill.dumps(get_element_name)
        >>> ft = PyEnvTask(func=func_str, args=[18], outputs=['element_name'])

        Of course, all references within the pickled object must be available
        at execution time. If the function has been pickled with 'dill',
        then 'dill' has to be available when unpickling.

        If 'py_hist_file' has been specified, then this task will produce
        a simple python file with the purpose to
        rerun the the call quickly outside of the FireWorks framework:

        >>> import pickle
        >>> func = pickle.loads(b'SOME_PICKLED_SEQUENCE')
        >>> args = [18]
        >>> kwargs = {}
        >>> output = func(*args, **kwargs)
    """

    _fw_name = 'PickledPyEnvTask'

    # try to unpickle with dill if 'func' is bytes
    def _get_func(self, fw_spec):
        """Get function from pickled bytes."""

        self._py_hist_append('import pickle')
        func_bytes = self['func']
        if isinstance(func_bytes, str):
            self.logger.info("type('func') == str, convert to bytes.")
            # when serializing a task, FireWorks apparently
            # puts the string representation
            # of the bytes objetc into the database
            func_bytes = eval(func_bytes)
        self._py_hist_append('func = pickle.loads({})'.format(func_bytes))
        func = pickle.loads(func_bytes)
        return func


class EvalPyEnvTask(PyEnvTask):
    """Same as PyEnvTask, but expects a lambda definition as string.

    Examples:

        >>> func_str = 'lambda x, y: x+y'
        >>> ft = EvalPyEnvTask(func=func_str, args=[1,2], outputs=['sum'])
        >>> fw_action = ft.run_task()

        returns an FWAction containing {'mod_spec': [{'_set': {'sum': 3}}]}.
    """

    _fw_name = 'EvalPyEnvTask'

    def _get_func(self, fw_spec):
        """Get function from string evaluation."""
        func_str = self['func']
        self._py_hist_append("func = eval('{:s}')".format(func_str))
        func = eval(func_str)
        return func
